# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains code to build a DeepLabV3.

Reference:
  - [Rethinking Atrous Convolution for Semantic Image Segmentation](
      https://arxiv.org/pdf/1706.05587.pdf)
"""
import tensorflow as tf

from deeplab2 import common
from deeplab2.model.decoder import aspp
from deeplab2.model.layers import convolutions

layers = tf.keras.layers


class DeepLabV3(layers.Layer):
    """A DeepLabV3 model.

    This model takes in features from an encoder and performs multi-scale context
    aggregation with the help of an ASPP layer. Finally, a classification head is
    used to predict a semantic segmentation.
    """

    def __init__(self,
                 decoder_options,
                 deeplabv3_options,
                 bn_layer=tf.keras.layers.BatchNormalization):
        """Creates a DeepLabV3 decoder of type layers.Layer.

        Args:
          decoder_options: Decoder options as defined in config_pb2.DecoderOptions.
          deeplabv3_options: Model options as defined in
            config_pb2.ModelOptions.DeeplabV3Options.
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
        """
        super(DeepLabV3, self).__init__(name='DeepLabV3')

        self._feature_name = decoder_options.feature_key
        self._aspp = aspp.ASPP(decoder_options.aspp_channels,
                               decoder_options.atrous_rates,
                               bn_layer=bn_layer)

        self._classifier_conv_bn_act = convolutions.Conv2DSame(
            decoder_options.decoder_channels,
            kernel_size=3,
            name='classifier_conv_bn_act',
            use_bias=False,
            use_bn=True,
            bn_layer=bn_layer,
            activation='relu')

        self._final_conv = convolutions.Conv2DSame(
            deeplabv3_options.num_classes, kernel_size=1, name='final_conv')

    def set_pool_size(self, pool_size):
        """Sets the pooling size of the ASPP pooling layer.

        Args:
          pool_size: A tuple specifying the pooling size of the ASPP pooling layer.
        """
        self._aspp.set_pool_size(pool_size)

    def get_pool_size(self):
        return self._aspp.get_pool_size()

    def reset_pooling_layer(self):
        """Resets the ASPP pooling layer to global average pooling."""
        self._aspp.reset_pooling_layer()

    def call(self, features, training=False):
        """Performs a forward pass.

        Args:
          features: A single input tf.Tensor or an input dict of tf.Tensor with
            shape [batch, height, width, channels]. If passed a dict, different keys
            should point to different features extracted by the encoder, e.g.
            low-level or high-level features.
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          A dictionary containing the semantic prediction under key
          common.PRED_SEMANTIC_LOGITS_KEY.
        """
        if isinstance(features, tf.Tensor):
            feature = features
        else:
            feature = features[self._feature_name]

        x = self._aspp(feature, training=training)

        x = self._classifier_conv_bn_act(x, training=training)

        return {common.PRED_SEMANTIC_LOGITS_KEY: self._final_conv(x)}

    @property
    def checkpoint_items(self):
        items = {
            common.CKPT_DEEPLABV3_ASPP: self._aspp,
            common.CKPT_DEEPLABV3_CLASSIFIER_CONV_BN_ACT:
                self._classifier_conv_bn_act,
            common.CKPT_SEMANTIC_LAST_LAYER: self._final_conv,
        }
        return items



"""
Geojson object oriented interface for region and site models.

This defines two classes ``SiteModel`` and ``RegionModel``, both of which
inherit from ``geojson.FeatureCollection``, so all geojson operations are
valid, but these classes contain extra convenience methods for loading,
dumping, manipulating, validating, and inspecting the data.

A non exhaustive list of convenience methods / properties of note are shared by
both site and region models are:

    * dumps - convert to a geojson string

    * pandas - convert to a geopandas data frame

    * coerce_multiple - read multiple geojson files at once.

    * header - a quick way to access the singular header row (region for region models and site for site models).

    * body_features - any row that is not a header is a body feature (site_summaries for region models and observations for site models).

    * validate - checks the site/region model against the schema.

    * random - classmethod to make a random instance of the site / region model for testing


Official T&E Terminology:

A Region Model gives an overview of entire region and summarizes all sites it contains. It consists of:

* A single header feature with type="region" that defines the region spacetime bounds

* Multiple body features with type="site_summary" that correspond to the bounds of an entire site. (i.e. there is one for each site in the region). A site summary has a "status" that applies to the entire temporal range of the site. (i.e. positive, negative, ignore)

A Site Model gives a detailed account of a single site within a region. It consists of:

* A single header feature with type="site" that roughly corresponds to one of the "site_summary" features in the region model. It also contains the holistic "status" field.

* Multiple body features with type="observation". This represents a single keyframe at a single point in time within the site's activity sequence. It contains a "current_phase" label that describes the specific phase of an activity at that current point in time.


Note: A site summary may exist on its own (i.e. without a corresponding site model) that gives a rough overview with holistic status, rough spatial bounds and a start / end date.


New region model specific convenience methods / properties are:

    * site_summaries
    * region_id
    * pandas_summaries
    * pandas_region

New site model specific convenience methods / properties are:

    * observations
    * pandas_observations
    * as_summary
    * region_id
    * site_id
    * status

SeeAlso:
    ../rc/registry.py

The following example illustrates how to read region / site models efficiently

Example:
    >>> # xdoctest: +REQUIRES(env:HAS_DVC)
    >>> import geowatch
    >>> dvc_data_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
    >>> region_models_dpath = dvc_data_dpath / 'annotations/drop6/region_models'
    >>> site_models_dpath = dvc_data_dpath / 'annotations/drop6/site_models'
    >>> from geowatch.geoannots import geomodels
    >>> region_models = list(geomodels.RegionModel.coerce_multiple(region_models_dpath))
    >>> site_models = list(geomodels.SiteModel.coerce_multiple(site_models_dpath, workers=8))
    >>> print(f'Number of region models: {len(region_models)}')
    >>> print(f'Number of site models: {len(site_models)}')
    >>> # Quick demo of associating sites to regions
    >>> region_id_to_sites = ub.group_items(site_models, key=lambda s: s.header['properties']['region_id'])
    >>> region_id_to_num_sites = ub.udict(region_id_to_sites).map_values(len)
    >>> print('region_id_to_num_sites = {}'.format(ub.urepr(region_id_to_num_sites, nl=1)))
    >>> # It is also easy to convert these models to geopandas
    >>> region_model = region_models[0]
    >>> gdf = region_model.pandas()
    >>> print(gdf)


XDEV_PROFILE=1 xdoctest ~/code/watch/geowatch/geoannots/geomodels.py


For testing the following example shows how to generate and inspect a random
site / region model.


Example:
    >>> from geowatch.geoannots.geomodels import *
    >>> # Generate a region model and also return its sites
    >>> region, sites = RegionModel.random(with_sites=True, rng=0)
    >>> # A region model consists of a region header
    >>> region_header = region.header
    >>> # And multiple site summaries. (We take the first one here)
    >>> site_summary = list(region.site_summaries())[0]
    >>> print('region_header.properties = {}'.format(ub.urepr(region_header['properties'], nl=1)))
    region_header.properties = {
        'type': 'region',
        'region_id': 'DR_R684',
        'version': '2.4.3',
        'mgrs': '51PXM',
        'start_date': '2011-05-28',
        'end_date': '2018-09-13',
        'originator': 'demo-truth',
        'model_content': 'annotation',
        'comments': 'demo-data',
    }
    >>> print('site_summary.properties = {}'.format(ub.urepr(site_summary['properties'], nl=1)))
    site_summary.properties = {
        'type': 'site_summary',
        'status': 'positive_annotated',
        'version': '2.0.1',
        'site_id': 'DR_R684_0000',
        'mgrs': '51PXM',
        'start_date': '2011-05-28',
        'end_date': '2018-09-13',
        'score': 1,
        'originator': 'demo',
        'model_content': 'annotation',
        'validated': 'True',
        'cache': {'color': [0.5511393746687864, 1.0, 0.0]},
    }
    >>> # A site model consists of a site header that roughly corresponds to a
    >>> # site summary in the region file
    >>> site = sites[0]
    >>> site_header = site.header
    >>> # It also contains one or more observations
    >>> site_obs = list(site.observations())[0]
    >>> print('site_header.properties = {}'.format(ub.urepr(site_header['properties'], nl=1)))
    site_header.properties = {
        'type': 'site',
        'status': 'positive_annotated',
        'version': '2.0.1',
        'site_id': 'DR_R684_0000',
        'mgrs': '51PXM',
        'start_date': '2011-05-28',
        'end_date': '2018-09-13',
        'score': 1,
        'originator': 'demo',
        'model_content': 'annotation',
        'validated': 'True',
        'cache': {'color': [0.5511393746687864, 1.0, 0.0]},
        'region_id': 'DR_R684',
    }
    >>> print('site_obs.properties = {}'.format(ub.urepr(site_obs['properties'], nl=1)))
    site_obs.properties = {
        'type': 'observation',
        'observation_date': '2011-05-28',
        'source': 'demosat-220110528T132754',
        'sensor_name': ...'demosat-2'...,
        'current_phase': 'No Activity',
        'is_occluded': 'False',
        'is_site_boundary': 'True',
        'score': 1.0,
    }


"""
import ubelt as ub
import geopandas as gpd
import geojson
import jsonschema
import copy
import json
from kwutil import util_time
from kwutil import util_progress

_VALID_SITE_OBSERVATION_FIELDS = {"type",
                                  "observation_date",
                                  "source",
                                  "sensor_name",
                                  "current_phase",
                                  "score",
                                  "misc_info",
                                  "is_occluded",
                                  "is_site_boundary"}


class _Model(ub.NiceRepr, geojson.FeatureCollection):
    """
    A base class for :class:`RegionModel` and :class:`SiteModel`.

    Note that because this extends :class:`geojson.FeatureCollection`, this is
    a dictionary.
    """
    type = 'FeatureCollection'
    _header_type = NotImplemented
    _body_type = NotImplemented

    def __nice__(self):
        return ub.urepr(self.info(), nl=2)

    def pandas(self):
        """
        Returns:
            geopandas.GeoDataFrame: the feature collection as data frame
        """
        gdf = gpd.GeoDataFrame.from_features(self['features'])
        return gdf

    def deepcopy(self):
        """
        Create a copy of this and all nested items

        Returns:
            Self
        """
        return copy.deepcopy(self)

    def dumps(self, **kwargs):
        """
        Serialize as json text

        Args:
            **kwargs: passed to :func:`json.dumps`

        Returns:
            str
        """
        return json.dumps(self, **kwargs)

    def dump(self, file, **kwargs):
        """
        Write json data to a file

        Args:
            file (PathLike | IO | None):
                Where to write the data. Can either be a path to a file or an
                open file pointer / stream.
            **kwargs: passed to :func:`json.dump`
        """
        _safer_dump_wrapper(json.dump, self, file, **kwargs)

    @classmethod
    def load(cls, file, **kwargs):
        """
        Load json data from a file pointer or path.

        Args:
            file (PathLike | IO | None):
                Where to read the data. Can either be a path to a file or an
                open file pointer / stream.
            **kwargs: passed to :func:`json.load`

        Returns:
            Self
        """
        data = _safer_load_wrapper(json.load, file, **kwargs)
        self = cls(**data)
        return self

    @classmethod
    def loads(cls, text, **kwargs):
        """
        Load json data from a string.

        Args:
            text (str): json text
            **kwargs: passed to :func:`json.loads`

        Returns:
            Self
        """
        data = json.loads(text, **kwargs)
        self = cls(**data)
        return self

    @classmethod
    def coerce_multiple_paths(cls, data):
        """
        Thin wrapper arround :func:`geowatch.util_gis.coerce_geojson_paths`
        """
        from kwgis.utils import util_gis
        return util_gis.coerce_geojson_paths(data)

    @classmethod
    def coerce_multiple(cls, data, allow_raw=False, workers=0, mode='thread',
                        verbose=1, desc=None, parse_float=None):
        """
        Load multiple geojson files.

        Args:
            arg (str | PathLike | List[str | PathLike]):
                an argument that is coerceable to one or more geojson files.

            **kwargs: see :func:`util_gis.coerce_geojson_datas`

        Yields:
            Self

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> import ubelt as ub
            >>> #
            >>> ### Setup demo data
            >>> dpath = ub.Path.appdir('geowatch/tests/geoannots/coerce_multiple')
            >>> dpath.delete().ensuredir()
            >>> regions, sites = [], []
            >>> for i in range(3):
            >>>     rm, sms = RegionModel.random(with_sites=True, rng=i)
            >>>     regions.append(rm)
            >>>     sites.extend(sms)
            >>> region_dpath = (dpath / 'region_models').ensuredir()
            >>> site_dpath = (dpath / 'site_models').ensuredir()
            >>> for region in regions:
            >>>     region_fpath = region_dpath / f'{region.region_id}.geojson'
            >>>     region_fpath.write_text(region.dumps())
            >>> for site in sites:
            >>>     site_fpath = site_dpath / f'{site.site_id}.geojson'
            >>>     site_fpath.write_text(site.dumps())
            >>> #
            >>> # Test coercing from a directory
            >>> regions2 = list(RegionModel.coerce_multiple(region_dpath))
            >>> sites2 = list(SiteModel.coerce_multiple(site_dpath))
            >>> assert len(regions2) == len(regions)
            >>> assert len(sites2) == len(sites)
            >>> #
            >>> # Test coercing from a glob pattern
            >>> regions3 = list(RegionModel.coerce_multiple(region_dpath / (regions[0].region_id + '*')))
            >>> sites3 = list(SiteModel.coerce_multiple(site_dpath / ('*.geojson')))
            >>> assert len(regions3) == 1
            >>> assert len(sites3) == len(sites)
            >>> #
            >>> # Test coercing from existing data
            >>> # Broken
            >>> # regions4 = list(RegionModel.coerce_multiple(regions))
            >>> # sites4 = list(SiteModel.coerce_multiple(sites))
            >>> # assert len(regions4) == len(regions)
            >>> # assert len(sites4) == len(sites)
        """
        from kwgis.utils import util_gis
        infos = list(util_gis.coerce_geojson_datas(
            data, format='json', allow_raw=allow_raw, workers=workers,
            mode=mode, verbose=verbose, desc=desc, parse_float=parse_float))
        for info in infos:
            item = cls(**info['data'])
            # Can we enrich each item with the path it was read from without
            # breaking dumps?
            # if 'fpath' in info:
            #     item.fpath = info['fpath']
            yield item

    @classmethod
    def coerce(cls, data, parse_float=None):
        """
        Coerce a :class:`RegionModel` or :class:`SiteModel` from some input.
        """
        import os
        if isinstance(data, cls):
            return data
        elif isinstance(data, dict):
            return cls.from_dict(data)
        elif isinstance(data, list):
            if all(isinstance(d, dict) and d['type'] == 'Feature' for d in data):
                return cls.from_features(data)
            else:
                raise TypeError('lists must a list of Features')
            return cls.from_dict(data)
        elif isinstance(data, gpd.GeoDataFrame):
            return cls.from_dataframe(data)
        elif isinstance(data, (str, os.PathLike)):
            got = list(cls.coerce_multiple(data, parse_float=parse_float, verbose=0))
            assert len(got) == 1
            return got[0]
        else:
            raise TypeError

    @classmethod
    def from_features(cls, features):
        """
        Args:
            gdf (GeoDataFrame):
        """
        self = cls(features=features)
        return self

    @classmethod
    def from_dataframe(cls, gdf):
        """
        Args:
            gdf (GeoDataFrame):
        """
        jsondct = gdf.__geo_interface__
        return cls(**jsondct)

    @classmethod
    def from_dict(cls, data):
        """
        Args:
            gdf (GeoDataFrame):
        """
        return cls(**data)

    @property
    def start_date(self):
        return util_time.coerce_datetime(self.header['properties']['start_date'])

    @property
    def end_date(self):
        return util_time.coerce_datetime(self.header['properties']['end_date'])

    @property
    def geometry(self):
        """
        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> RegionModel.random().geometry
            >>> SiteModel.random().geometry
        """
        from shapely import geometry
        return geometry.shape(self.header['geometry'])

    def load_schema(self, strict=True):
        raise NotImplementedError('abstract')

    def body_features(self):
        """
        Iterate over non-header items in the geojson feature list
        """
        for feat in self['features']:
            prop = feat['properties']
            if prop['type'] == self._body_type:
                yield feat

    def strip_body_features(self):
        """
        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> self = RegionModel.random()
            >>> assert len(list(self.body_features())) > 0
            >>> self.strip_body_features()
            >>> assert len(list(self.body_features())) == 0
        """
        self['features'] = [self.header]

    @property
    def header(self):
        """
        Get the single feature representing the header
        """
        for feat in self['features']:
            prop = feat['properties']
            if prop['type'] == self._header_type:
                return feat

    def _validate_quick_checks(self):
        header = self.header
        if header is None:
            raise AssertionError('Geo Model has no header')

        if header is not self.features[0]:
            import warnings
            warnings.warn('Header should be the first feature')

        if header['properties']['type'] != self._header_type:
            raise AssertionError('Header type is wrong')

        if self['type'] != 'FeatureCollection':
            raise AssertionError('GeoModels should be FeatureCollections')

        feature_types = ub.dict_hist([
            f['properties']['type'] for f in self.features])
        assert feature_types.pop(self._header_type, 0) == 1, 'Missing header'
        assert set(feature_types).issubset({self._body_type}), f'Unexpected feature types: {feature_types}'

        start_date = self.start_date
        end_date = self.end_date
        if start_date is not None and end_date is not None:
            if end_date < start_date:
                raise AssertionError(f'bad date: start_date={start_date} end_date={end_date}')

    def _validate_schema(self, strict=True, verbose=1, parts=True):
        schema = self.load_schema(strict=strict)
        try:
            jsonschema.validate(self, schema)
        except jsonschema.ValidationError as _full_ex:
            full_ex = _full_ex
            if verbose:
                print(f'self={self}')
                _report_jsonschema_error(full_ex)
            if parts:
                try:
                    self._validate_parts(strict=strict, verbose=verbose)
                except Exception as _part_ex:
                    part_ex = _part_ex
                    part_ex.full_ex = full_ex
                    raise part_ex
            raise full_ex

    def validate(self, strict=True, verbose=1, parts=True):
        """
        Validates that the model conforms to its schema and does a decent job
        of localizing where errors are.

        Args:
            strict (bool):
                if False, SMART-specific fields have their restrictions
                loosened. Defaults to True.

            verbose (bool):
                if True prints out extra information on an errors

            parts (bool):
                if True, attempts to determine what part of the data is causing
                the error.
        """
        self._validate_quick_checks()
        self._validate_schema(strict=strict, verbose=verbose, parts=parts)

    def _validate_parts(self, strict=True, verbose=1):
        """
        Runs jsonschema validation checks on each part of the feature
        collection independently to better localize where the errors are.

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> self = RegionModel.random(rng=0)
            >>> self._validate_parts(strict=False)
            >>> self = SiteModel.random(rng=0)
            >>> self._validate_parts(strict=False)
        """
        import jsonschema
        schema = ub.udict(self.load_schema(strict=strict))
        schema - {'properties', 'required', 'title', 'type'}
        defs = schema[chr(36) + 'defs']
        header_schema = schema | (defs[self._header_type + '_feature'])
        body_schema = schema | (defs[self._body_type + '_feature'])
        try:
            jsonschema.validate(self.header, header_schema)
        except jsonschema.ValidationError as e:
            if verbose:
                _report_jsonschema_error(e)
            raise
        for obs_feature in self.body_features():
            try:
                jsonschema.validate(obs_feature, body_schema)
            except jsonschema.ValidationError as e:
                if verbose:
                    _report_jsonschema_error(e)
                raise

    def _update_cache_key(self):
        """
        Ensure we are using the up to date schema cache.

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> self = RegionModel.random(rng=0)
            >>> feat = list(self.site_summaries())[0]
            >>> self._update_cache_key()
            >>> assert 'annotation_cache' not in feat['properties']
            >>> feat['properties']['annotation_cache'] = {'foo': 'bar'}
            >>> self._update_cache_key()
            >>> # An old cache key, updates the new one.
            >>> assert 'cache' in feat['properties']
            >>> assert feat['properties']['cache']['foo'] == 'bar'
            >>> # But it wont overwrite.
            >>> feat['properties']['annotation_cache'] = {'foo': 'baz'}
            >>> self._update_cache_key()
            >>> assert 'cache' in feat['properties']
            >>> assert feat['properties']['cache']['foo'] == 'bar'

            self.header['properties']['cache'] = None
            self.fixup()
            self.validate(strict=0)
            assert self.header['properties']['cache'] == {}
        """
        for feat in self['features']:
            prop = feat['properties']
            _update_propery_cache(prop)

    def ensure_isodates(self):
        """
        Ensure that dates are provided as dates and not datetimes

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> region = RegionModel.random()
            >>> region.header['properties']['start_date'] = '1970-01-01T000000'
            >>> region.ensure_isodates()
            >>> assert region.header['properties']['start_date'] == '1970-01-01'
        """
        date_keys = ['start_date', 'end_date']
        for feat in self['features']:
            props = feat['properties']
            for key in date_keys:
                if key in props:
                    oldval = props[key]
                    if oldval is not None:
                        dt = util_time.coerce_datetime(oldval)
                        try:
                            newval = dt.date().isoformat()
                        except Exception:
                            print('ERROR: oldval = {}'.format(ub.urepr(oldval, nl=1)))
                        props[key] = newval

    def fix_backwards_dates(self):
        """
        If start and end dates are backwards, flip them.

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> ss = SiteSummary.random()
            >>> ss['properties']['start_date'] = '1970-01-01T000000'
            >>> ss.ensure_isodates()
            >>> assert ss['properties']['start_date'] == '1970-01-01'
        """
        start_date = self.start_date
        end_date = self.end_date
        if start_date is not None and end_date is not None:
            if end_date < start_date:
                _new_start = self.header['properties']['end_date']
                _new_end = self.header['properties']['start_date']
                self.header['properties']['start_date'] = _new_start
                self.header['properties']['end_date'] = _new_end

    @property
    def model_type(self):
        return self.header['properties']['type']

    @property
    def model_id(self):
        header_id_key = self._header_type + '_id'
        return self.header['properties'][header_id_key]


def _report_jsonschema_error(ex):
    import rich

    def print_validation_error_info(ex, depth=1):
        if ex.parent is not None:
            max_depth = print_validation_error_info(ex.parent, depth=depth + 1)
        else:
            max_depth = depth
        rich.print(f'[yellow] error depth = {depth} / {max_depth}')
        print('ex.__dict__ = {}'.format(ub.urepr(ex.__dict__, nl=3)))
        return depth

    rich.print('[red] JSON VALIDATION ERROR')
    print_validation_error_info(ex)
    # ub.IndexableWalker(self)[ex.absolute_path]
    # ub.IndexableWalker(schema)[ex.schema_path]
    rich.print(ub.codeblock(
        '''
        [yellow] jsonschema validation notes:
            * depsite our efforts, information to debug the issue may not be shown, double check your schema and instance manually.
            * anyOf schemas may print the error, and not the part you intended to match.
            * oneOf schemas may not explicitly say that you matched both.
        '''))
    rich.print('[red] JSON VALIDATION ERROR')


class RegionModel(_Model):
    """
    Wrapper around a geojson region model FeatureCollection

    TODO:
        Rename to Region?

    Example:
        >>> from geowatch.geoannots.geomodels import *  # NOQA
        >>> self = RegionModel.random()
        >>> print(self)
        >>> self.validate(strict=False)
    """
    _header_type = 'region'
    _body_type = 'site_summary'

    def info(self):
        header = self.header
        prop = '<no region header>' if header is None else header['properties']
        info = {
            'num_site_summaries': len(list(self.site_summaries())),
            'properties': prop,
        }
        return info

    @classmethod
    def load_schema(cls, strict=True):
        import geowatch
        schema = geowatch.rc.registry.load_region_model_schema(strict=strict)
        return schema

    def site_summaries(self):
        yield from (SiteSummary(**f) for f in self.body_features())

    @classmethod
    def coerce(cls, data, parse_float=None):
        """
        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> import ubelt as ub
            >>> dpath = ub.Path.appdir('geowatch/tests/geoannots/coerce').ensuredir()
            >>> region = RegionModel.random(with_sites=False, rng=0)
            >>> data = fpath = (dpath/ 'region.geojson')
            >>> fpath.write_text(region.dumps())
            >>> region_models = list(RegionModel.coerce_multiple(fpath))
            >>> region_model = RegionModel.coerce(fpath)
        """
        self = super().coerce(data, parse_float=parse_float)
        assert self.header['properties']['type'] == 'region'
        return self

    def pandas_summaries(self):
        """
        Returns:
            geopandas.GeoDataFrame: the site summaries as a data frame

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> self = RegionModel.random()
            >>> gdf = self.pandas_summaries()
            >>> print(gdf)
            >>> # Test empty pandas summary
            >>> self = RegionModel.random(num_sites=0)
            >>> gdf = self.pandas_summaries()
            >>> print(gdf)
            >>> assert len(gdf) == 0
        """
        from kwgis.utils import util_gis
        crs84 = util_gis.get_crs84()
        site_summaries = list(self.site_summaries())
        if len(site_summaries):
            gdf = gpd.GeoDataFrame.from_features(site_summaries, crs=crs84)
        else:
            # TODO: could infer more columns here.
            default_sitesum_columns = [
                'geometry', 'type', 'status', 'version', 'site_id', 'mgrs',
                'start_date', 'end_date', 'score', 'originator', 'model_content',
            ]
            gdf = gpd.GeoDataFrame.from_features(
                [], crs=crs84, columns=default_sitesum_columns)

        return gdf

    def pandas_region(self):
        """
        Returns:
            geopandas.GeoDataFrame: the region header as a data frame

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> self = RegionModel.random()
            >>> print(self.pandas_region())
        """
        from kwgis.utils import util_gis
        crs84 = util_gis.get_crs84()
        gdf = gpd.GeoDataFrame.from_features([self.header], crs=crs84)
        return gdf

    pandas_header = pandas_region
    pandas_body = pandas_summaries

    @classmethod
    def random(cls, with_sites=False, **kwargs):
        """
        Creates a random region model optionally with random sites for use in
        testing / demos.

        Args:
            with_sites (bool):
                also returns site models if True

            **kwargs :
                passed to
                :func:`geowatch.demo.metrics_demo.demo_truth.random_region_model`.
                Some of these args are:
                    num_sites
                    num_observations
                    start_time
                    end_time
                    region_poly
                    rng

        Returns:
            RegionModel | Tuple[RegionModel, SiteModelCollection]

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> region1 = RegionModel.random(with_sites=False, rng=0)
            >>> region2, sites2 = RegionModel.random(with_sites=True, rng=0)
            >>> assert region1 == region2, 'rngs should be the same'
        """
        from geowatch.demo.metrics_demo import demo_truth

        region, sites, _ = demo_truth.random_region_model(
            **kwargs, with_renderables=False)

        region = cls(**region)

        if with_sites:
            sites = SiteModelCollection([SiteModel(**s) for s in sites])
            return region, sites
        else:
            return region

    def add_site_summary(self, summary):
        """
        Add a site summary to the region.

        Args:
            summary (SiteSummary | SiteModel):
                a site summary or site model. If given as a site model
                it is converted to a site summary and then added.

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> region = RegionModel.random(num_sites=False)
            >>> site1 = SiteModel.random(region=region)
            >>> site2 = SiteModel.random(region=region)
            >>> site3 = SiteModel.random(region=region)
            >>> summary = site2.as_summary()
            >>> region.add_site_summary(site1)
            >>> region.add_site_summary(summary)
            >>> region.add_site_summary(dict(site3.as_summary()))
            >>> import pytest
            >>> with pytest.raises(TypeError):
            ...     region.add_site_summary(dict(site3))
            >>> assert len(list(region.site_summaries())) == 3
        """
        if isinstance(summary, SiteModel):
            summary = summary.as_summary()
        if summary['type'] != 'Feature' or summary['properties']['type'] != 'site_summary':
            raise TypeError('Input was not a site summary or coercable type')
        self['features'].append(summary)

    @property
    def region_id(self):
        """
        Get the region_id from the geojson header
        """
        return self.header['properties']['region_id']

    def fixup(self):
        """
        Fix common issues with this region model

        Returns:
            RegionModel
        """
        self._update_cache_key()
        self.remove_invalid_properties()
        self.ensure_isodates()
        self.fix_backwards_dates()
        self.fix_multipolygons()
        return self

    def fix_multipolygons(self):
        from shapely import from_geojson
        from shapely.geometry import MultiPolygon, mapping

        for sitesum in self.body_features():
            geom = from_geojson(json.dumps(sitesum['geometry']))

            if isinstance(geom, MultiPolygon):
                convex_hull = geom.buffer(0).convex_hull
                # Have to serialize/deserialize here as shapely
                # returns tuple coordinates which might throw off
                # downstream code (expecting lists)
                new_geom = json.loads(json.dumps(mapping(convex_hull)))

                sitesum['geometry'] = new_geom

    def remove_invalid_properties(self):
        """
        Remove invalid properties from this region model that have caused
        issues in the past.
        """
        props = self.header['properties']
        bad_region_header_properties = ['validated', 'score', 'site_id', 'status', 'socre']
        for key in bad_region_header_properties:
            props.pop(key, None)

        bad_sitesum_keys = ['region_id', 'validate', 'validated',
                            'predicted_phase_transition',
                            'predicted_phase_transition_date']
        non_nullable_sitesum_keys = ['score']
        for sitesum in self.body_features():
            siteprops = sitesum['properties']
            for key in bad_sitesum_keys:
                siteprops.pop(key, None)
            for key in non_nullable_sitesum_keys:
                if key in siteprops and siteprops[key] is None:
                    siteprops.pop(key, None)

    def ensure_comments(self):
        props = self.header['properties']

        props['comments'] = props.get('comments', '')

    def infer_header(self, region_header=None):
        """
        Infer any missing header information from site summaries.

        If this region model does not have a header, but it contains site
        summaries, then use that information to infer a header value. Useful
        when converting site summaries to full region models.

        Args:
            region_header (RegionHeader):
                if specified, use this information when possible and then
                infer the rest.

        SeeAlso:

            * :func:`SiteModelCollection.as_region_model`

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> # Make a region without a header
            >>> self = RegionModel.random()
            >>> self.features.remove(self.header)
            >>> assert self.header is None
            >>> # Infer the header using site summaries
            >>> self.infer_header()
            >>> assert self.header is not None
        """
        current_header = self.header

        if region_header is not None:
            if current_header is not None:
                raise ValueError('cannot specify a region header when one already exists')
            region_header = RegionHeader.coerce(region_header)
            region_header = copy.deepcopy(region_header)
        else:
            if current_header is not None:
                region_header = current_header
            else:
                region_header = RegionHeader.empty()

        site_summaries = list(self.site_summaries())
        region_header = _infer_region_header_from_site_summaries(
            region_header, site_summaries)

        if region_header is not current_header:
            assert current_header is None
            self.features.insert(0, region_header)


class SiteModel(_Model):
    """
    Wrapper around a geojson site model FeatureCollection

    TODO:
        Rename to Site?

    Example:
        >>> from geowatch.geoannots.geomodels import *  # NOQA
        >>> self = SiteModel.random()
        >>> print(self)
        >>> self.validate(strict=False)
    """
    _header_type = 'site'
    _body_type = 'observation'

    def info(self):
        header = self.header
        prop = '<no site header>' if header is None else header['properties']
        info = {}
        info['num_observations'] = len(list(self.observations()))
        if header is not None:
            info['header_geom_type'] = header['geometry']['type']
        info['properties'] = prop
        return info

    @classmethod
    def load_schema(cls, strict=True):
        import geowatch
        schema = geowatch.rc.registry.load_site_model_schema(strict=strict)
        return schema

    @property
    def header(self):
        for feat in self['features']:
            prop = feat['properties']
            if prop['type'] == 'site':
                return feat

    def observations(self):
        """
        Features containing specific observations with phase labels
        """
        yield from self.body_features()

    def pandas_observations(self):
        """
        Returns:
            geopandas.GeoDataFrame: the site summaries as a data frame

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> self = SiteModel.random()
            >>> gdf = self.pandas_observations()
            >>> print(gdf)
            >>> # Test empty pandas summary
            >>> del self.features[1:]
            >>> gdf = self.pandas_observations()
            >>> print(gdf)
            >>> assert len(gdf) == 0
        """
        from kwgis.utils import util_gis
        crs84 = util_gis.get_crs84()
        features = list(self.observations())
        if len(features):
            gdf = gpd.GeoDataFrame.from_features(features, crs=crs84)
        else:
            gdf = gpd.GeoDataFrame.from_features(features, crs=crs84,
                                                 columns=['geometry'])
        return gdf

    def pandas_site(self):
        """
        Returns:
            geopandas.GeoDataFrame: the region header as a data frame

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> self = SiteModel.random()
            >>> print(self.pandas_site())
        """
        from kwgis.utils import util_gis
        crs84 = util_gis.get_crs84()
        gdf = gpd.GeoDataFrame.from_features([self.header], crs=crs84)
        return gdf

    pandas_header = pandas_site
    pandas_body = pandas_observations

    @classmethod
    def random(cls, rng=None, region=None, site_poly=None, **kwargs):
        """
        Args:
            rng (int | str | RandomState | None) :
                seed or random number generator

            region (RegionModel | None):
                if specified generate a new site in this region model.
                (This will overwrite some of the kwargs).

            site_poly (kwimage.Polygon | shapely.geometry.Polygon | None):
                if specified, this polygon is used as the geometry for new site
                models. Note: all site models will get this geometry, so
                typically this is only used when num_sites=1.

            **kwargs :
                passed to :func:`geowatch.demo.metrics_demo.demo_truth.random_region_model`.

        Returns:
            SiteModel

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> region1 = RegionModel.random(with_sites=False, rng=0)
            >>> region2, sites2 = RegionModel.random(with_sites=True, rng=0)
            >>> assert region1 == region2, 'rngs should be the same'

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> region = RegionModel.random(with_sites=False, rng=0)
            >>> site = SiteModel.random(region=region)
            >>> assert region.region_id == site.region_id

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> import kwimage
            >>> region = RegionModel.random(with_sites=False, rng=0)
            >>> # Test specification of the site geometry.
            >>> site_poly = kwimage.Polygon.coerce(region.geometry)
            >>> site = SiteModel.random(region=region, site_poly=site_poly)
            >>> assert abs(region.geometry.area - site.geometry.area) < 1e-7
            >>> site = SiteModel.random(region=region, site_poly=site_poly.scale(10))
            >>> assert abs(region.geometry.area - site.geometry.area) > 1e-7
        """
        from geowatch.demo.metrics_demo import demo_truth
        kwargs.setdefault('with_renderables', False)
        kwargs['site_poly'] = site_poly
        if region is not None:
            kwargs['region_poly'] = region.header.geometry
            kwargs['region_id'] = region.region_id
        _, sites, _ = demo_truth.random_region_model(num_sites=1, rng=rng, **kwargs)
        return cls(**sites[0])

    def as_summary(self):
        """
        Modify and return this site header feature as a site-summary body
        feature for a region model.

        Returns:
            SiteSummary
        """
        header = self.header
        if header is None:
            raise IndexError('Site model has no header')
        else:
            header = SiteHeader(**header)
            summary = header.as_summary()
            return SiteSummary(**summary)

    @property
    def region_id(self):
        return self.header['properties']['region_id']

    @property
    def site_id(self):
        return self.header['properties']['site_id']

    @property
    def status(self):
        return self.header['properties']['status']

    def fix_geom(self):
        from shapely.geometry import shape
        from shapely.validation import make_valid
        from shapely.geometry import MultiPolygon
        for feat in self.features:
            geom = shape(feat['geometry'])
            if geom.geom_type in {'MultiPolygon', 'Polygon'}:
                make_valid(geom)
            else:
                geom = geom.buffer(3).convex_hull
                geom = MultiPolygon([geom])
            feat['geometry'] = geom.__geo_interface__

    def fix_sensor_names(self):
        for feat in self.observations():
            prop = feat['properties']
            if prop.get('sensor_name') == 'WorldView 1':
                prop['sensor_name'] = 'WorldView'

    def fix_current_phase_salient(self):
        for feat in self.observations():
            prop = feat['properties']
            current_phase = prop.get('current_phase')
            if current_phase is not None:
                if 'salient' in current_phase:
                    prop['current_phase'] = prop['current_phase'].replace(
                        'salient', 'Active Construction')

    def fix_multipolygons(self):
        from shapely import from_geojson
        from shapely.geometry import MultiPolygon, mapping

        site = self.header
        geom = from_geojson(json.dumps(site['geometry']))

        if isinstance(geom, MultiPolygon):
            convex_hull = geom.buffer(0).convex_hull
            # Have to serialize/deserialize here as shapely
            # returns tuple coordinates which might throw off
            # downstream code (expecting lists)
            new_geom = json.loads(json.dumps(mapping(convex_hull)))

            site['geometry'] = new_geom

    def fixup(self):
        """
        Fix common issues with this site model

        Returns:
            SiteModel
        """
        self._update_cache_key()
        self.clamp_scores()
        self.fix_sensor_names()
        self.fix_aliased_properties()
        self.ensure_isodates()
        self.fix_current_phase_salient()
        self.fix_backwards_dates()
        self.fix_old_schema_properties()
        self.fix_multipolygons()
        # self.fix_geom()
        return self

    def fix_aliased_properties(self):
        """
        Some models are written with aliased properties (e.g. stop_date instead
        of end_date). This fixes them.
        """
        feat = self.header
        props = feat['properties']
        end_date = props.get('end_date', None)
        if end_date is None:
            if 'stop_date' in props:
                props['end_date'] = props.pop('stop_date')

    def fix_old_schema_properties(self):
        """
        If an old schema property exists and is not null, move it to
        the cache.
        """
        old_keys = ['comments']
        for feat in self.features:
            props = feat['properties']
            for key in old_keys:
                if key in props:
                    old_value = props.pop(key)
                    if old_value is not None:
                        if 'cache' not in props:
                            props['cache'] = {}
                        # Dont overwrite an existing key with the same name
                        # in this case we just drop the bad value
                        if key not in props['cache']:
                            props['cache'][key] = old_value

    def ensure_isodates(self):
        """
        Ensure that dates are provided as dates and not datetimes

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> site = SiteModel.random()
            >>> # Set props as datetimes
            >>> site.header['properties']['start_date'] = '1970-01-01T000000'
            >>> site.features[1]['properties']['observation_date'] = '1970-01-01T000000'
            >>> site.ensure_isodates()
            >>> # The fixup ensure dates
            >>> assert site.features[1]['properties']['observation_date'] == '1970-01-01'
            >>> assert site.header['properties']['start_date'] == '1970-01-01'
        """
        date_keys = ['start_date', 'end_date', 'predicted_phase_transition_date']
        feat = self.header
        props = feat['properties']
        for key in date_keys:
            if key in props:
                old_val = props[key]
                if old_val is not None:
                    props[key] = util_time.coerce_datetime(old_val).date().isoformat()

        date_keys = ['observation_date']
        for feat in self.body_features():
            props = feat['properties']
            for key in date_keys:
                if key in props:
                    old_val = props[key]
                    if old_val is not None:
                        props[key] = util_time.coerce_datetime(old_val).date().isoformat()

    def clamp_scores(self):
        for feat in self.features:
            fprop = feat['properties']
            old_score = fprop.get('score', None)
            if old_score is not None:
                fprop['score'] = float(max(min(1.0, fprop['score']), 0.0))

    def remove_invalid_properties(self):
        """
        Remove invalid properties from this site model
        """
        # T&E site schema no longer allows extraneous keys to be
        # included in region / site models; removing all unsupported
        # keys (could consider putting in 'misc_info' rather than
        # deleting, though not clear if 'misc_info' will be supported
        # in the future)
        for obs in self.observations:
            oprop = obs['properties']

            to_remove = set()
            for k in oprop.keys():
                if k not in _VALID_SITE_OBSERVATION_FIELDS:
                    to_remove.add(k)

            for k in to_remove:
                del oprop[k]

    def _manual_validation(self):
        """
        Hard coded checks. The jsonschema is pretty bad at identifing where
        errors are, so these are some hard coded checks that hit some simple
        errors we have seen before.
        """
        features = self.features
        if len(features) < 2:
            raise AssertionError('should have at least two features')

        type_to_expected_fields = {
            'feature': {
                'required': {'type', 'properties', 'geometry'},
                'optional': set(),
            },
            'site': {
                'required': {
                    'type', 'site_id', 'region_id', 'version', 'mgrs', 'model_content',
                    'start_date', 'end_date', 'status', 'originator'},
                'optional': {
                    'misc_info', 'validated', 'score',
                    'predicted_phase_transition_date',
                    'predicted_phase_transition'
                }
            },
            'observation': {
                'required': {
                    'type', 'observation_date', 'source', 'sensor_name',
                    'current_phase', 'is_occluded', 'is_site_boundary'
                },
                'optional': {
                    'misc_info', 'score',
                }
            }
        }

        type_to_expected_geoms = {
            'site': {'Polygon'},
            'observation': {'Polygon', 'MultiPolygon'},
        }

        def check_expected_fields(have, type):
            expected = type_to_expected_fields[type]
            missing = expected['required'] - have
            extra = have - (expected['required'] | expected['optional'])
            if extra:
                yield {
                    'msg': f'Extra fields: {extra}'
                }
            if missing:
                yield {
                    'msg': f'Missing fields: {missing}'
                }
            return errors

        def check_expected_geom(geom, type):
            allowed_types = type_to_expected_geoms[type]
            if geom.geom_type not in allowed_types:
                yield {
                    'msg': f'{type} must be in {allowed_types}: got {geom.geom_type}'
                }

        from shapely.geometry import shape
        errors = []
        for feat in features:
            have = set(feat.keys())
            errors += list(check_expected_fields(have, type='feature'))
            geom = shape(feat['geometry'])
            props = feat['properties']
            proptype = props['type']
            if proptype == 'site':
                have = set(props.keys())
                errors += list(check_expected_fields(have, type='site'))
                errors += list(check_expected_geom(geom, type='site'))
            elif proptype == 'observation':
                have = set(props.keys())
                errors += list(check_expected_fields(have, type='observation'))
                errors += list(check_expected_geom(geom, type='observation'))
            else:
                errors += {
                    'msg': f'Unknown site type: {proptype}',
                }

        if len(errors):
            print('errors = {}'.format(ub.urepr(errors, nl=1)))
            raise AssertionError


class _Feature(ub.NiceRepr, geojson.Feature):
    """
    Base class for geojson features that conform to an IARPA geomodel spec

    Example:
        >>> # Test the class variables for subclasses are defined correctly
        >>> assert RegionHeader._feat_type == 'region'
        >>> assert SiteSummary._feat_type == 'site_summary'
        >>> assert SiteHeader._feat_type == 'site'
        >>> assert Observation._feat_type == 'observation'
        >>> assert RegionHeader._model_cls is RegionModel
        >>> assert SiteSummary._model_cls is RegionModel
        >>> assert SiteHeader._model_cls is SiteModel
        >>> assert Observation._model_cls is SiteModel
    """
    type = 'Feature'
    _model_cls = NotImplemented
    _feat_type = NotImplemented

    def __nice__(self):
        return ub.urepr(self.info(), nl=2)

    def info(self):
        info = {
            'properties': self['properties'],
        }
        return info

    @property
    def properties(self):
        return self['properties']

    @classmethod
    def load_schema(cls, strict=True):
        """
        Return the sub-schema for the approprite header / body feature
        based on the declaration of _model_cls and _feat_type
        """
        assert cls._model_cls is not NotImplemented
        assert cls._feat_type is not NotImplemented
        region_schema = cls._model_cls.load_schema(strict=strict)
        schema = ub.udict(region_schema)
        schema - {'properties', 'required', 'title', 'type'}
        defs = schema[chr(36) + 'defs']
        feat_schema = schema | (defs[cls._feat_type + '_feature'])
        return feat_schema

    def validate(self, strict=True, verbose=1):
        """
        Validate this sub-schema
        """
        feat_schema = self.load_schema(strict=strict)
        try:
            jsonschema.validate(self, feat_schema)
        except jsonschema.ValidationError as e:
            if verbose:
                _report_jsonschema_error(e)
            raise

    def ensure_isodates(self):
        """
        Ensure that dates are provided as dates and not datetimes

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> ss = SiteSummary.random()
            >>> ss['properties']['start_date'] = '1970-01-01T000000'
            >>> ss.ensure_isodates()
            >>> assert ss['properties']['start_date'] == '1970-01-01'
        """
        date_keys = ['start_date', 'end_date']
        props = self['properties']
        for key in date_keys:
            if key in props:
                oldval = props[key]
                if oldval is not None:
                    dt = util_time.coerce_datetime(oldval)
                    try:
                        newval = dt.date().isoformat()
                    except Exception:
                        print('ERROR: oldval = {}'.format(ub.urepr(oldval, nl=1)))
                    props[key] = newval

    def infer_mgrs(self, strict=True):
        """

        Args:
            strict (bool): if False, do not error if this fails

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> ss = SiteSummary.random()
            >>> ss.infer_mgrs()
        """

        from shapely.geometry import shape
        import mgrs
        if self.geometry is None:
            handle_error('Cannot infer mgrs, missing geometry',
                         extype=Exception, strict=strict)
        else:
            _geom = shape(self.geometry)
            lon = _geom.centroid.xy[0][0]
            lat = _geom.centroid.xy[1][0]
            mgrs_code = mgrs.MGRS().toMGRS(lat, lon, MGRSPrecision=0)
            self.properties['mgrs'] = mgrs_code
        return self

    def _fix_geojson_geometry(self):
        """
        Turn any shapely objects into serializable geojson
        """
        geometry =  self['geometry']
        if not isinstance(geometry, dict):
            import geojson
            self['geometry'] = geojson.loads(geojson.dumps(geometry))


class Point(_Feature):
    ...


class _SiteOrSummaryMixin:
    """
    Site summaries and site headers are nearly the same
    """

    # Data for conversion between site / site-summaries
    _cache_keys = {
        'site_summary': 'annotation_cache',
        'site': 'misc_info',
    }
    # Record non-common properties between the two similar schemas
    _only_properties = {
        'site_summary': [
            'comments'
        ],
        'site': [
            'predicted_phase_transition_date',
            'predicted_phase_transition',
            'region_id',
        ]
    }

    @property
    def start_date(self):
        return util_time.coerce_datetime(self['properties']['start_date'])

    @property
    def end_date(self):
        return util_time.coerce_datetime(self['properties']['end_date'])

    @property
    def site_id(self):
        return self['properties']['site_id']

    @property
    def status(self):
        return self['properties']['status']

    def to_point(self):
        date = self.start_date
        from shapely import geometry
        # import json
        # import shapely
        geom = geometry.shape(self['geometry'])
        point_geom = geom.centroid
        # point_geom = json.loads(shapely.to_geojson(geom.centroid))

        point = Point(
            properties={'site_id': self.site_id,
                        'base_version': '2.0.2',
                        'status': self.status,
                        'date': date.date().isoformat(),
                        'type': 'point',
                        'point_version': '1.0',
                        'date_version': '1.1000000000000001'},
            geometry=point_geom)
        return point

    def _update_cache_key(self):
        """
        Ensure we are using the up to date schema cache.
        """
        prop = self['properties']
        _update_propery_cache(prop)

    def _convert(self, new_cls):
        """
        Common logic for converting site <-> site_summary

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> site = SiteModel.random()
            >>> site.validate(strict=False)
            >>> region = RegionModel.random()
            >>> region.validate(strict=False)
            >>> site1 = SiteHeader(**site.header)
            >>> site1.validate(strict=False)
            >>> summary1 = SiteSummary(**ub.peek(region.body_features()))
            >>> summary1.validate(strict=False)
            >>> summary2 = site1.as_summary()
            >>> summary2.validate(strict=False)
            >>> import pytest
            >>> with pytest.raises(Exception):
            >>>     site2 = summary1.as_site()
            >>> summary1['properties']['cache']['region_id'] = region.region_id
            >>> site2 = summary1.as_site()
            >>> site2.validate(strict=False)
            >>> # Check the round-trip conversion
            >>> summary3 = site2.as_summary()
            >>> site3 = summary2.as_site()
            >>> summary1_ = SiteSummary(**summary1.copy())
            >>> summary1_._update_cache_key()
            >>> site1_ = SiteHeader(**site1.copy())
            >>> site1_._update_cache_key()
            >>> assert summary3 == summary1_ and summary3 is not summary1
            >>> assert site3 == site1_ and site3 is not site1
            >>> # Revalidate everything to ensure no memory issues happened
            >>> summary3.validate(strict=0)
            >>> summary2.validate(strict=0)
            >>> summary1.validate(strict=0)
            >>> site3.validate(strict=0)
            >>> site2.validate(strict=0)
            >>> site1.validate(strict=0)
            >>> site.validate(strict=0)
            >>> region.validate(strict=0)
        """
        old_type = self._feat_type
        new_type = new_cls._feat_type
        old_cache_key = self._cache_keys[old_type]
        old_only_props = self._only_properties[old_type]
        new_cache_key  = self._cache_keys[new_type]
        new_only_props  = self._only_properties[new_type]

        feat = self.copy()
        props = feat['properties'].copy()

        if 1:
            # Use new scheme
            _update_propery_cache(props)
            old_cache_key = 'cache'
            new_cache_key = 'cache'

        feat['properties'] = props
        assert props['type'] == old_type
        props['type'] = new_type
        if old_cache_key in props:
            props[new_cache_key] = props.pop(old_cache_key)
        cache = props.get(new_cache_key, {})
        for key in new_only_props:
            if key in cache:
                props[key] = cache.pop(key)
        for key in old_only_props:
            if key in props:
                cache[key] = props.pop(key)
        if cache:
            props[new_cache_key] = cache

        if old_type == 'site_summary':
            if 'region_id' not in props:
                raise Exception(ub.paragraph(
                    '''
                    Cannot convert a site-summary to a site header when the
                    region-id is unknown. As a workaround you can set the
                    .properties.annotation_cache.region_id
                    '''))

        new = new_cls(**feat)
        return new


class RegionHeader(_Feature):
    """
    The region header feature of a region model.
    """
    _model_cls = RegionModel
    _feat_type = RegionModel._header_type

    @classmethod
    def empty(cls):
        """
        Create an empty region header
        """
        self = cls(
            properties={
                "type": "region",
                "region_id": None,
                "version": "2.4.3",
                "mgrs": None,
                "start_date": None,
                "end_date": None,
                "originator": None,
                "model_content": None,
            },
            geometry=None,
        )
        return self

    @classmethod
    def coerce(cls, data):
        """
        Example:
            >>> data = RegionModel.random()
            >>> h1 = RegionHeader.coerce(data)
            >>> h2 = RegionHeader.coerce(data.header)
            >>> assert h1 == h2

            RegionHeader.coerce(orig_region_model)
        """
        if isinstance(data, cls):
            return data
        if isinstance(data, dict):
            if 'type' in data:
                if data['type'] == 'Feature':
                    return cls(**data)
                elif data['type'] == 'FeatureCollection':
                    return cls(**RegionModel(**data).header)
        raise TypeError(data)

    def ensure_isodates(self):
        date_keys = ['start_date', 'end_date', 'predicted_phase_transition_date']
        feat = self
        props = feat['properties']
        for key in date_keys:
            if key in props:
                old_val = props[key]
                if old_val is not None:
                    props[key] = util_time.coerce_datetime(old_val).date().isoformat()
        return self


class SiteSummary(_Feature, _SiteOrSummaryMixin):
    """
    The site-summary body feature of a region model.
    """
    _model_cls = RegionModel
    _feat_type = RegionModel._body_type

    @classmethod
    def from_geopandas_frame(cls, df, drop_id=True):
        json_text = df.to_json(drop_id=drop_id)
        json_data = json.loads(json_text)
        for feat in json_data['features']:
            if feat['properties']['type'] == 'site_summary':
                yield cls(**feat)

    def as_site(self):
        """
        Modify and return this site summary feature as a site header feature
        for a site model.

        Returns:
            SiteHeader

        Example:
            >>> # Convert a RegionModel to a collection of SiteModels
            >>> from geowatch.geoannots import geomodels
            >>> region = geomodels.RegionModel.random()
            >>> sites = []
            >>> for sitesum in region.site_summaries():
            >>>     # Current hacky way to pass along region ids
            >>>     sitesum['properties']['cache']['region_id'] = region.region_id
            >>>     # This only produces a site header, we may need to add
            >>>     # observations to the site model itself as well
            >>>     site_header = sitesum.as_site()
            >>>     site = SiteModel(features=[site_header])
            >>>     sites.append(site)
        """
        new_cls = SiteHeader
        return self._convert(new_cls)

    def fixup(self):
        """
        Fixup the site summary
        """
        self._update_cache_key()
        self._remove_invalid_properties()
        # self.ensure_isodates()
        return self

    def _remove_invalid_properties(self):
        """
        Remove invalid properties from this region model that have caused
        issues in the past.
        """
        bad_sitesum_keys = ['region_id', 'validate', 'validated',
                            'predicted_phase_transition',
                            'predicted_phase_transition_date']
        non_nullable_sitesum_keys = ['score']
        siteprops = self['properties']
        for key in bad_sitesum_keys:
            siteprops.pop(key, None)
        for key in non_nullable_sitesum_keys:
            if key in siteprops and siteprops[key] is None:
                siteprops.pop(key, None)

    @classmethod
    def coerce(cls, data):
        if isinstance(data, cls):
            self = data
        elif isinstance(data, dict):
            assert data['type'] == 'Feature'
            assert data['properties']['type'] == 'site_summary'
            self = cls(**data)
        else:
            raise TypeError(type(data))
        return self

    @classmethod
    def random(cls, rng=None, region=None, site_poly=None, **kwargs):
        """
        Args:
            rng (int | str | RandomState | None) :
                seed or random number generator

            region (RegionModel | None):
                if specified generate a new site in this region model.
                (This will overwrite some of the kwargs).

            site_poly (kwimage.Polygon | shapely.geometry.Polygon | None):
                if specified, this polygon is used as the geometry for new site
                models. Note: all site models will get this geometry, so
                typically this is only used when num_sites=1.

            **kwargs :
                passed to :func:`geowatch.demo.metrics_demo.demo_truth.random_region_model`.

        Returns:
            SiteSummary

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> sitesum = SiteSummary.random(rng=0)
            >>> print('sitesum = {}'.format(ub.urepr(sitesum, nl=2)))
        """
        site = SiteModel.random(rng=rng, region=region, site_poly=site_poly, **kwargs)
        return site.as_summary()


class SiteHeader(_Feature, _SiteOrSummaryMixin):
    """
    The site header feature of a site model.
    """
    _model_cls = SiteModel
    _feat_type = SiteModel._header_type

    @classmethod
    def empty(cls):
        """
        Create an empty region header

        Example:
            from geowatch.geoannots.geomodels import *  # NOQA
            self = SiteHeader.empty()
            ...
        """
        self = cls(
            properties={
                "type": "site",
                "version": "2.4.3",
                "mgrs": None,
                "status": None,
                "model_content": None,
                "score": None,
                "start_date": None,
                "end_date": None,
                "originator": None,
                "validated": 'False',
            },
            geometry=None,
        )
        return self

    def as_summary(self):
        """
        Modify and return this site header feature as a site-summary body
        feature for a region model.

        Returns:
            SiteSummary
        """
        new_cls = SiteSummary
        return self._convert(new_cls)

    @classmethod
    def coerce(cls, data):
        if isinstance(data, cls):
            self = data
        elif isinstance(data, dict):
            assert data['type'] == 'Feature'
            assert data['properties']['type'] == 'site'
            self = cls(**data)
        else:
            raise TypeError(type(data))
        return self

    @classmethod
    def random(cls, rng=None, region=None, site_poly=None, **kwargs):
        """
        Args:
            rng (int | str | RandomState | None) :
                seed or random number generator

            region (RegionModel | None):
                if specified generate a new site header in this region model.

            site_poly (kwimage.Polygon | shapely.geometry.Polygon | None):
                if specified, this polygon is used as the geometry

            **kwargs :
                passed to :func:`geowatch.demo.metrics_demo.demo_truth.random_region_model`.

        Returns:
            SiteHeader

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> site_header = SiteHeader.random(rng=0)
            >>> print('site_header = {}'.format(ub.urepr(site_header, nl=2)))
        """
        site = SiteModel.random(rng=rng, region=region, site_poly=site_poly, **kwargs)
        return site.header


class Observation(_Feature):
    """
    The observation body feature of a site model.

    Example:
        >>> from geowatch.geoannots.geomodels import *  # NOQA
        >>> Observation()
    """
    _model_cls = SiteModel
    _feat_type = SiteModel._body_type

    @classmethod
    def coerce(cls, data):
        if isinstance(data, cls):
            self = data
        elif isinstance(data, dict):
            assert data['type'] == 'Feature'
            assert data['properties']['type'] == 'observation'
            self = cls(**data)
        else:
            raise TypeError(type(data))
        return self

    @property
    def observation_date(self):
        return util_time.coerce_datetime(self['properties']['observation_date'])

    @classmethod
    def random(cls, rng=None, region=None, site_poly=None, **kwargs):
        """
        Args:
            rng (int | str | RandomState | None) :
                seed or random number generator

            region (RegionModel | None):
                if specified generate a new observation in this region model.

            site_poly (kwimage.Polygon | shapely.geometry.Polygon | None):
                if specified, this polygon is used as the geometry for new observation

            **kwargs :
                passed to :func:`geowatch.demo.metrics_demo.demo_truth.random_region_model`.

        Returns:
            Observation

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> obs = Observation.random(rng=0)
            >>> print(f'obs={obs}')
        """
        site = SiteModel.random(rng=rng, region=region, site_poly=site_poly, **kwargs)
        return list(site.observations())[0]

    @classmethod
    def empty(cls):
        """
        Create an empty observation

        Example:
            >>> from geowatch.geoannots.geomodels import *  # NOQA
            >>> self = Observation.empty()
            >>> print(f'self = {ub.urepr(self, nl=2)}')
        """
        self = cls(
            properties={
                'type': 'observation',
                'observation_date': None,  # e.g. '2011-05-28',
                'source': None,  # e.g. 'demosat-220110528T132754',
                'sensor_name': None,  # e.g. 'demosat-2',
                'current_phase': None,  # e.g. "No Activity".
                'is_occluded': None,  # quirk / note: bool should be a string
                'is_site_boundary': None,  # quirk / note: bool should be a string
                'score': None,
            },
            geometry=None,
        )
        return self


# def _site_header_from_observations(observations, mgrs_code, site_id, status, summary_geom=None):
#     """
#     Consolodate site observations into a site header
#     """
#     if summary_geom is None:
#         summary_geom = unary_union(
#             [kwimage.MultiPolygon.coerce(o["geometry"]).to_shapely() for o in observations]
#         ).convex_hull
#     start_date = observations[0]["properties"]["observation_date"]
#     end_date = observations[-1]["properties"]["observation_date"]
#     sitesum_props = {
#         "type": "site_summary",
#         "status": status,
#         "version": "2.0.1",
#         "site_id": site_id,
#         "mgrs": mgrs_code,
#         "start_date": start_date,
#         "end_date": end_date,
#         "score": 1,
#         "originator": "demo",
#         "model_content": "annotation",
#         "validated": "True",
#     }
#     site_summary = geojson.Feature(
#         properties=sitesum_props,
#         geometry=kwimage.Polygon.coerce(summary_geom).to_geojson(),
#     )
#     return site_summary


class ModelCollection(list):
    """
    A storage container for multiple site / region models
    """

    def fixup(self):
        pman = util_progress.ProgressManager()
        with pman:
            for model in pman.progiter(self, desc='fixup'):
                model.fixup()
        return self

    def validate(self, strict=False, stop_on_failure=True, verbose=1, mode='process', workers=0):
        """
        Validate multiple models in parallel
        """
        import rich
        # pman = util_progress.ProgressManager(backend='progiter')
        pman = util_progress.ProgressManager()
        with pman:
            jobs = ub.JobPool(mode='process', max_workers=workers)
            for model in pman.progiter(self, desc='submit validate models'):
                job = jobs.submit(model.validate, strict=strict, verbose=verbose)
                job.model = model
            num_passed = 0
            errors = []
            prog = pman.progiter(jobs.as_completed(), total=len(jobs), desc='collect validate models')
            for job in prog:
                try:
                    job.result()
                except Exception as ex:
                    rich.print(f'[red] ERROR: failed to validate {job.model.model_id} : {job.model.model_type} in a collection')
                    errors.append((ex, job.model))
                    prog.set_extra(f'Passed: {num_passed}, Failed: {len(errors)}')
                    if stop_on_failure:
                        raise
                else:
                    num_passed += 1
                prog.set_extra(f'Passed: {num_passed}, Failed: {len(errors)}')
            if errors:
                num_failed = len(errors)
                num_total = len(jobs)
                failed_model_ids = [model.model_id for ex, model in errors]
                rich.print(f'[red] ERROR: failed to validate {num_failed} / {num_total} models')
                rich.print('failed_model_ids = {}'.format(ub.urepr(failed_model_ids, nl=1)))
                raise Exception(f'Failed to validate {num_failed} / {num_total} models')


class PointModelCollection(ModelCollection):
    ...


class SiteModelCollection(ModelCollection):

    def as_region_model(self, region_header=None, region_id=None, strict=True):
        """
        Convert a set of site models to a region model

        Args:
            region (RegonModel | RegionHeader | None):
                If specified, use this information to generate the new region
                header. If unspecified, we attempt to infer this from the site
                models.

            region_id (str | None):
                if specified, use this as the region id

            strict (bool):
                if False, ignore missing uninferable information.

        Returns:
            RegonModel: a new region model where each site in this collection
                appears as a site summary.

        Example:
            >>> from geowatch.geoannots.geomodels import RegionModel
            >>> region, sites = RegionModel.random(with_sites=True, rng=0)
            >>> self = SiteModelCollection(sites)
            >>> self.as_region_model()
        """
        site_summaries = [s.as_summary() for s in self]
        site_header_properties = [site.header['properties'] for site in self]

        if region_header is not None:
            region_header = RegionHeader.coerce(region_header)
            region_header = copy.deepcopy(region_header)
        else:
            region_header = RegionHeader.empty()

        if region_id is not None:
            region_header['properties']['region_id'] = region_id

        region_props = region_header['properties']
        # note: region_id does not appear in a site summary, but it does in the
        # site model.
        key = 'region_id'
        if region_props.get(key, None) is None:
            if len(self) == 0:
                handle_error(f'No sites. Unable to infer {key}.', strict=strict)
            else:
                region_props[key] = _rectify_keys(key, site_header_properties)

        region_header = _infer_region_header_from_site_summaries(
            region_header, site_summaries, strict)

        region_features = [region_header] + site_summaries
        region_model = RegionModel(features=region_features)
        return region_model

    def to_point_model(self):
        points = [s.as_summary().to_point() for s in self]
        point_model = PointModel(points)
        return point_model

    @classmethod
    def coerce(cls, data):
        """
        Create a collection of site models from input - usually a directory
        containing site model geojson files.

        SeeAlso:
            :func:`SiteModel.coerce_multiple`.
        """
        return cls(SiteModel.coerce_multiple(data))


class PointModel(_Model):
    ...


def _infer_region_header_from_site_summaries(region_header, site_summaries, strict=True):
    """
    Given a RegionHeader use site_summaries to fill missing data.
    """
    if region_header is None:
        region_header = RegionHeader.empty()
    region_props = region_header.get('properties', None)

    if region_props.get('type', None) is None:
        region_props['type'] = 'region'

    site_summary_properties = [sitesum['properties'] for sitesum in site_summaries]

    shared_unique_properties = ['originator', 'model_content', 'mgrs']

    for key in shared_unique_properties:
        if region_props.get(key, None) is None:
            try:
                if len(site_summaries) == 0:
                    handle_error(f'No sites. Unable to infer {key}.', strict=strict)
                else:
                    region_props[key] = _rectify_keys(key, site_summary_properties)
            except ValueError as ex:
                # Allow MGRS to fail. We can use region geometry to get the
                # right one.
                if key != 'mgrs':
                    if strict:
                        raise
                    else:
                        print(f'Warning: ex = {ub.urepr(ex, nl=1)}')

    if region_props.get('start_date', None) is None:
        if len(site_summaries) == 0:
            handle_error('No sites. Unable to infer start_date.', strict=strict)
        dates = [p['start_date'] for p in site_summary_properties if p['start_date'] is not None]
        if len(dates) == 0:
            handle_error('No sites with start dates', strict=strict)
        else:
            region_props['start_date'] = min(dates)

    if region_props.get('end_date', None) is None:
        if len(site_summaries) == 0:
            handle_error('No sites. Unable to infer end_date.', strict=strict)
        dates = [p['end_date'] for p in site_summary_properties if p['end_date'] is not None]
        if len(dates) == 0:
            handle_error('No sites with end dates', strict=strict)
        else:
            region_props['end_date'] = max(dates)

    if region_header.get('geometry', None) is None:
        if len(site_summaries) == 0:
            handle_error(f'No sites. Unable to infer {key}.', strict=strict)
            # region_header['geometry'] = {'type': 'Point', 'coordinates': []}
        else:
            from shapely.ops import unary_union
            import kwimage
            import shapely.geometry
            site_geoms = [shapely.geometry.shape(s['geometry']).buffer(0)
                          for s in site_summaries]
            sh_geom = unary_union(site_geoms).envelope
            dct_geom = kwimage.Polygon.from_shapely(sh_geom).to_geojson()
            region_header['geometry'] = dct_geom

    if region_props.get('mgrs', None) is None:
        RegionHeader(**region_header).infer_mgrs(strict=strict)

    return region_header


def _rectify_keys(key, properties_list):
    """
    Given a key and a list of dictionaries, extract the value for that key in
    all dictionaries and check they are all the same.

    Args:
        key (str): key of interest
        properties_list (List[Dict[str, T]]): multiple property dictionaries

    Returns:
        T: value from properties dictionaries.
    """
    if len(properties_list) == 0:
        raise ValueError(f'No sites. Unable to infer {key}.')
    unique_values = ub.dict_hist(p[key] for p in properties_list)
    if len(unique_values) > 1:
        msg = (f'More than one key={key!r} in with unique_values={unique_values!r}')
        print(msg)
        raise ValueError(msg)
    value = list(unique_values)[0]
    return value


def handle_error(msg, extype=ValueError, strict=True):
    import rich
    if strict:
        raise extype(msg)
    else:
        rich.print(f'[yellow]WARNING: {msg}')


def _update_propery_cache(prop):
    """
    Move to the new cache schema
    """
    if 'annotation_cache' in prop or 'misc_info' in prop:
        cache = prop.get('cache', {})
        cache = ub.udict.union(prop.pop('annotation_cache', {}), cache)
        cache = ub.udict.union(prop.pop('misc_info', {}), cache)
        if cache:
            prop['cache'] = cache
    if 'cache' in prop:
        if prop['cache'] is None:
            prop['cache'] = {}


def coerce_site_or_region_model(model_data):
    """
    Args:
        model_data (dict): A geojson FeatureCollection that should correspond
            to a SiteModel or RegionModel.

    Returns:
        SiteModel | RegionModel - return type depends on the input data
    """
    assert isinstance(model_data, dict)
    assert model_data['type'] == 'FeatureCollection'
    for feat in model_data['features']:
        assert feat['type'] == 'Feature'
        if feat['properties']['type'] == 'region':
            return RegionModel(**model_data)
        elif feat['properties']['type'] == 'site':
            return SiteModel(**model_data)
    raise AssertionError('Did not find a region or site header')


def _safer_dump_wrapper(dump_func, obj, file, mode='w', temp_file='auto', **kwargs):
    """
    Helper to write a "dump" function that takes a file pointer or file path.

    TODO:
        Reduce code duplication. Refactor this into a shared utility and use in
        this and kwcoco.CocoDataset.dump - which is where this code was derived
        from. Does this go in kwutil or just vendor it where needed?

    Args:
        dump_func (Callable):
            The dump function to wrap (e.g. json.dump).

        obj (object): the object to pass to the dump func

        file (PathLike | IO | None):
            Where to write the data. Can either be a path to a file or an

        temp_file (bool | str):
            Argument to :func:`safer.open`.  Ignored if ``file`` is not a
            PathLike object. Defaults to 'auto', which is False on Windows
            and True everywhere else.

        **kwargs Arguments to the file-based "dump" function.
    """
    import os

    try:
        fpath = os.fspath(file)
    except TypeError:
        input_was_pathlike = False
    else:
        input_was_pathlike = True

    if input_was_pathlike:
        import safer
        if temp_file == 'auto':
            temp_file = not ub.WIN32

        with safer.open(fpath, mode=mode, temp_file=temp_file) as fp:
            dump_func(obj, fp, **kwargs)
    else:
        # We are likely dumping to a real file.
        dump_func(obj, file, **kwargs)


def _safer_load_wrapper(load_func, file, mode='r', temp_file='auto', **kwargs):
    """
    Helper to write a "load" function that takes a file pointer or file path.
    """
    import os

    try:
        fpath = os.fspath(file)
    except TypeError:
        input_was_pathlike = False
    else:
        input_was_pathlike = True

    if input_was_pathlike:
        import safer
        if temp_file == 'auto':
            temp_file = not ub.WIN32

        with safer.open(fpath, mode=mode, temp_file=temp_file) as fp:
            return load_func(fp, **kwargs)
    else:
        # We are likely dumping to a real file.
        return load_func(file, **kwargs)



# flake8: noqa
import numpy as np

import torch


class TensorPoly(object):
    def __init__(self, pos, poly_slice, batch, batch_size, is_endpoint=None):
        """

        :param pos:
        :param poly_slice:
        :param batch: Batch index for each node
        :param is_endpoint: One value per node. If true, that node is an endpoint and is thus part of an open polyline
        """
        assert pos.shape[0] == batch.shape[0]
        self.pos = pos
        self.poly_slice = poly_slice
        self.batch = batch
        self.batch_size = batch_size
        self.is_endpoint = is_endpoint

        self.to_padded_index = None  # No pad initially
        self.to_unpadded_poly_slice = None

    @property
    def num_nodes(self):
        return self.pos.shape[0]

    def to(self, device):
        self.pos = self.pos.to(device)
        self.poly_slice = self.poly_slice.to(device)
        self.batch = self.batch.to(device)
        if self.is_endpoint is not None:
            self.is_endpoint = self.is_endpoint.to(device)
        if self.to_padded_index is not None:
            self.to_padded_index = self.to_padded_index.to(device)
        if self.to_unpadded_poly_slice is not None:
            self.to_unpadded_poly_slice = self.to_unpadded_poly_slice.to(device)


def polygons_to_tensorpoly(polygons_batch):
    """
    Parametrizes N polygons into a 1d grid to be used in 1d conv later on:
    - pos (n1+n2+..., 2) concatenation of polygons vertex positions
    - poly_slice (poly_count, 2) polygon vertex slices [start, end)

    :param polygons_batch: Batch of polygons: [[(n1, 2), (n2, 2), ...], ...]
    :return: TensorPoly(pos, poly_slice, batch, batch_size, is_endpoint)
    """
    # TODO: If there are no polygons
    batch_size = len(polygons_batch)
    is_endpoint_list = []
    batch_list = []
    polygon_list = []
    for i, polygons in enumerate(polygons_batch):
        for polygon in polygons:
            if not np.max(np.abs(polygon[0] - polygon[-1])) < 1e-6:
                # Polygon is open
                is_endpoint = np.zeros(polygon.shape[0], dtype=np.bool)
                is_endpoint[0] = True
                is_endpoint[-1] = True
            else:
                # Polygon is closed, remove last redundant point
                polygon = polygon[:-1, :]
                is_endpoint = np.zeros(polygon.shape[0], dtype=np.bool)
            batch = i * np.ones(polygon.shape[0], dtype=np.long)
            is_endpoint_list.append(is_endpoint)
            batch_list.append(batch)
            polygon_list.append(polygon)
    pos = np.concatenate(polygon_list, axis=0)
    is_endpoint = np.concatenate(is_endpoint_list, axis=0)
    batch = np.concatenate(batch_list, axis=0)

    slice_start = 0
    poly_slice = np.empty((len(polygon_list), 2), dtype=np.long)
    for i, polygon in enumerate(polygon_list):
        slice_end = slice_start + polygon.shape[0]
        poly_slice[i][0] = slice_start
        poly_slice[i][1] = slice_end
        slice_start = slice_end
    pos = torch.tensor(pos, dtype=torch.float)
    is_endpoint = torch.tensor(is_endpoint, dtype=torch.bool)
    poly_slice = torch.tensor(poly_slice, dtype=torch.long)
    batch = torch.tensor(batch, dtype=torch.long)
    tensorpoly = TensorPoly(pos=pos, poly_slice=poly_slice, batch=batch, batch_size=batch_size, is_endpoint=is_endpoint)
    return tensorpoly


def _get_to_padded_index(poly_slice, node_count, padding):
    """
    Pad each polygon with a cyclic padding scheme on both sides.
    Increases length of x by (padding[0] + padding[1])*polygon_count values.

    :param poly_slice:
    :param padding:
    :return:
    """
    assert len(poly_slice.shape) == 2, "poly_slice should have shape (poly_count, 2), not {}".format(poly_slice.shape)
    poly_count = poly_slice.shape[0]
    range_tensor = torch.arange(node_count, device=poly_slice.device)
    to_padded_index = torch.empty((node_count + (padding[0] + padding[1]) * poly_count, ), dtype=torch.long, device=poly_slice.device)
    to_unpadded_poly_slice = torch.empty_like(poly_slice)
    start = 0
    for poly_i in range(poly_count):
        poly_indices = range_tensor[poly_slice[poly_i, 0]:poly_slice[poly_i, 1]]

        # Repeat poly_indices if necessary when padding exceeds polygon length
        vertex_count = poly_indices.shape[0]
        left_repeats = padding[0] // vertex_count
        left_padding_remaining = padding[0] % vertex_count
        right_repeats = padding[1] // vertex_count
        right_padding_remaining = padding[1] % vertex_count
        total_repeats = left_repeats + right_repeats
        poly_indices = poly_indices.repeat(total_repeats + 1)  # +1 includes the original polygon

        if left_padding_remaining:
            poly_indices = torch.cat([poly_indices[-left_padding_remaining:],
                                      poly_indices,
                                      poly_indices[:right_padding_remaining]])
        else:
            poly_indices = torch.cat([poly_indices, poly_indices[:right_padding_remaining]])

        end = start + poly_indices.shape[0]
        to_padded_index[start:end] = poly_indices
        to_unpadded_poly_slice[poly_i, 0] = start  # Init value
        to_unpadded_poly_slice[poly_i, 1] = end  # Init value
        start = end
    to_unpadded_poly_slice[:, 0] += padding[0]  # Shift all inited values to the right one
    to_unpadded_poly_slice[:, 1] -= padding[1]  # Shift all inited values to the right one
    return to_padded_index, to_unpadded_poly_slice


def tensorpoly_pad(tensorpoly, padding):
    to_padded_index, to_unpadded_poly_slice = _get_to_padded_index(tensorpoly.poly_slice, tensorpoly.num_nodes, padding)
    tensorpoly.to_padded_index = to_padded_index
    tensorpoly.to_unpadded_poly_slice = to_unpadded_poly_slice
    return tensorpoly


def main():
    device = "cuda"

    np.random.seed(0)
    torch.manual_seed(0)
    padding = (0, 1)

    batch_size = 2
    poly_count = 3
    vertex_min_count = 4
    vertex_max_count = 5

    polygons_batch = []
    for batch_i in range(batch_size):
        polygons = []
        for poly_i in range(poly_count):
            vertex_count = np.random.randint(vertex_min_count, vertex_max_count)
            polygon = np.random.uniform(0, 1, (vertex_count, 2))
            polygons.append(polygon)
        polygons_batch.append(polygons)
    print("polygons_batch:")
    print(polygons_batch)
    tensorpoly = polygons_to_tensorpoly(polygons_batch)
    print("batch:")
    print(tensorpoly.batch)
    print("pos:")
    print(tensorpoly.pos.shape)
    print("poly_slice:")
    print(tensorpoly.poly_slice.shape)
    print(tensorpoly.poly_slice)

    tensorpoly.to(device)

    tensorpoly = tensorpoly_pad(tensorpoly, padding)
    to_padded_index = tensorpoly.to_padded_index
    print("to_padded_index:")
    print(to_padded_index.shape)
    print(to_padded_index)


if __name__ == "__main__":
    main()



class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    DEBUG = '\033[31;40m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def print_info(*args):
    print(bcolors.OKBLUE + " ".join(map(str, args)) + bcolors.ENDC)


def print_success(*args):
    print(bcolors.OKGREEN + " ".join(map(str, args)) + bcolors.ENDC)


def print_failure(*args):
    print(bcolors.FAIL + " ".join(map(str, args)) + bcolors.ENDC)


def print_error(*args):
    print_failure(*args)


def print_warning(*args):
    print(bcolors.WARNING + " ".join(map(str, args)) + bcolors.ENDC)


def print_debug(*args):
    print(bcolors.DEBUG + " ".join(map(str, args)) + bcolors.ENDC)


def print_format_table():
    """
    prints table of formatted text format options
    """
    for style in range(8):
        for fg in range(30, 38):
            s1 = ''
            for bg in range(40, 48):
                format = ';'.join([str(style), str(fg), str(bg)])
                s1 += '\x1b[%sm %s \x1b[0m' % (format, format)
            print(s1)
        print('\n')


def main():
    print_format_table()

    print_info("Info")
    print_success("Success")
    print_failure("Failure")
    print_error("ERROR")
    print_warning("WARNING")
    print_debug("Debug")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import sys
import traceback
import shutil

import scriptconfig as scfg
import ubelt as ub

"""
Notes:

    A quick local test to see if the tensorboard dependencies are acting up.

    # Set this to the docker image you want to test.
    IMAGE=watch:0.7.4-95c6b4b2-strict-pyenv3.11.2-20230623T134259-0400-from-01080c26

    docker run -it $IMAGE python -c "if 1:
        import numpy as np
        import geowatch
        import ubelt as ub
        from geowatch.tasks.depth_pcd.model import getModel
        model = getModel()
        expt_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_expt', hardware='auto')
        model.load_weights(expt_dvc_dpath + '/models/depth_pcd/basicModel2.h5')
        out = model.predict(np.zeros((1,400,400,3)))
        shapes = [o.shape for o in out]
        print('shapes = {}'.format(ub.urepr(shapes, nl=1)))
    "

"""


class DzyneParallelSiteValiConfig(scfg.DataConfig):
    """
    Run DZYNE's parallel site validation framework component

    python ~/code/watch/geowatch/cli/smartflow/run_dzyne_parallel_site_vali.py
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
        '''
        Path to the STAC items this step can use as inputs.
        This is usually an S3 Path.
        '''), alias=['input_stac_path'])

    input_region_path = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
        '''
        Path to input T&E Baseline Framework Region definition JSON
        '''))

    output_path = scfg.Value(None, type=str, position=3, required=True, help=ub.paragraph(
        '''
        Path to the STAC items that register the outputs of this stage.
        This is usually an S3 Path.
        '''), alias=['output_stac_path'])

    # depth_model_fpath = scfg.Value("/models/depthPCD/basicModel2.h5", type=str, position=4, required=True, help='path to depth model weights')

    aws_profile = scfg.Value(None, help=ub.paragraph(
        '''
        AWS Profile to use for AWS S3 CLI commands
        '''))

    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='Run AWS CLI commands with --dryrun flag')

    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
        '''
        S3 Output directory for STAC item / asset egress
        '''))

    depth_score_config = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        Raw json/yaml or a path to a json/yaml file that specifies the
        config for depth scorer.
        '''))

    depth_filter_config = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        Raw json/yaml or a path to a json/yaml file that specifies the
        config for the depth filter.
        '''))

    skip_on_fail = scfg.Value(False, help=ub.paragraph(
        '''
        If an error occurs, pass through input region / sites unchanged.
        '''))

    input_region_models_asset_name = scfg.Value('cropped_region_models_bas', type=str, required=False, help=ub.paragraph(
        '''
        Which region model assets to use as input
        '''))

    input_site_models_asset_name = scfg.Value('cropped_site_models_bas', type=str, required=False, help=ub.paragraph(
        '''
        Which site model assets to to use as input
        '''))


def main():
    config = DzyneParallelSiteValiConfig.cli(strict=True)
    print('config = {}'.format(ub.urepr(config, nl=1, align=':')))
    run_dzyne_parallel_site_vali_for_baseline(config)


def run_dzyne_parallel_site_vali_for_baseline(config):
    """
    SeeAlso:
        ~/code/watch/geowatch/tasks/depth_pcd/score_tracks.py
        ~/code/watch/geowatch/tasks/depth_pcd/filter_tracks.py
    """
    from geowatch.cli.smartflow_ingress import smartflow_ingress
    from geowatch.cli.smartflow_egress import smartflow_egress
    # from geowatch.cli.concat_kwcoco_videos import concat_kwcoco_datasets
    from geowatch.utils.util_framework import download_region, determine_region_id
    # from geowatch.tasks.fusion.predict import predict
    # from kwutil.util_yaml import Yaml

    input_path = config.input_path
    input_region_path = config.input_region_path
    output_path = config.output_path
    outbucket = config.outbucket
    aws_profile = config.aws_profile
    dryrun = config.dryrun

    ####
    # DEBUGGING:
    # Print info about what version of the code we are running on
    from geowatch.utils.util_framework import NodeStateDebugger
    node_state = NodeStateDebugger()
    node_state.print_environment()
    node_state.print_local_invocation(config)

    # 1. Ingress data
    print("* Running baseline framework kwcoco ingress *")
    ingress_dir = ub.Path('/tmp/ingress')
    ingressed_assets = smartflow_ingress(
        input_path=input_path,
        assets=[
            'cropped_kwcoco_for_sv',
            'cropped_kwcoco_for_sv_assets',
            config.input_site_models_asset_name,
            config.input_region_models_asset_name,
        ],
        outdir=ingress_dir,
        aws_profile=aws_profile,
        dryrun=dryrun
    )

    # # 2. Download and prune region file
    print("* Downloading and pruning region file *")

    # CHECKME: Is this the region model with site summaries from the tracker?
    local_region_path = '/tmp/region.json'

    download_region(
        input_region_path=input_region_path,
        output_region_path=local_region_path,
        aws_profile=aws_profile,
        strip_nonregions=True,
        ensure_comments=True,
    )

    # Determine the region_id in the region file.
    region_id = determine_region_id(local_region_path)

    # 3. Run the Site Validation Filter
    print("* Running the Site Validation Filter *")
    from geowatch.tasks.depth_pcd import score_tracks
    from geowatch.tasks.depth_pcd import filter_tracks
    from kwutil.util_yaml import Yaml

    default_score_config = ub.udict({
        'model_fpath': None,
    })
    score_config = (default_score_config
                    | Yaml.coerce(config.depth_score_config or {}))
    if score_config.get('model_fpath', None) is None:
        raise ValueError('Requires model_fpath')

    filter_config = (Yaml.coerce(config.depth_filter_config or {}))

    # TODO: The input / output site and region paths should be specified as
    # parameters passed to us by the DAG.
    input_kwcoco_fpath = ub.Path(ingressed_assets['cropped_kwcoco_for_sv'])
    input_sites_dpath = ub.Path(ingressed_assets[config.input_site_models_asset_name])
    input_region_dpath = ub.Path(ingressed_assets[config.input_region_models_asset_name])
    input_region_fpath = input_region_dpath / f'{region_id}.geojson'
    # input_region_fpath = local_region_path  # is this right?

    scored_kwcoco_fpath = ingress_dir / "poly_depth_scored.kwcoco.zip"

    output_site_manifest_fpath = ingress_dir / "depth_filtered_sites_manifest.json"
    output_sites_dpath = ingress_dir / "depth_filtered_sites"
    output_region_dpath = ingress_dir / "depth_filtered_regions"
    output_region_fpath = output_region_dpath / f'{region_id}.geojson'

    output_region_dpath.ensuredir()
    output_sites_dpath.ensuredir()

    # 3.3. Check that we have at least one "video" (BAS identified
    # site) to run over; if not skip SV fusion and KWCOCO to GeoJSON
    import kwcoco
    input_coco_dset = kwcoco.CocoDataset(input_kwcoco_fpath)
    print('input_coco_dset = {}'.format(ub.urepr(input_coco_dset, nl=1)))
    num_videos = input_coco_dset.n_videos
    print(f'num_videos={num_videos}')

    if num_videos == 0:
        # Copy input region model into region_models outdir to be updated
        # (rather than generated from tracking, which may not have the
        # same bounds as the original)

        # Not sure if the above case is the right comment, but leaving this
        # here to guarentee the region with site summaries is passed forward
        # TODO: the dino code should just be robust to this.
        input_region_fpath.copy(output_region_fpath)
    else:
        # 3.4 Run DZYNE depth_pcd
        print("* Running DZYNE depth_pcd *")

        try:
            cmdline = 0
            score_kwargs = dict(
                **score_config,
                # Should be the SV-cropped kwcoco file that contains start and ending
                # high resolution images where videos correspond to proposed sites for
                # this region.
                input_kwcoco=input_kwcoco_fpath,

                # Should be the region models containing the current site summaries
                # from the previous step.
                input_region=input_region_fpath,

                # This is a kwcoco file used internally in this step where scores
                # are assigned to each track. The next step will use this.
                out_kwcoco=scored_kwcoco_fpath,
            )
            score_tracks.main(cmdline=cmdline, **score_kwargs)

            cmdline = 0
            filter_kwargs = {
                **filter_config,

                # The kwcoco file contining depth scores that this step will use to
                # filter the input sites / site summaries.
                'input_kwcoco': scored_kwcoco_fpath,

                # Should be the region models containing the current site summaries
                # from the previous step.
                'input_region': input_region_fpath,

                # Should be the folder containing all of the sites corresponding to the
                # sites in the input_region
                'input_sites': input_sites_dpath,

                # The output region model to be used by the next step
                'output_region_fpath': output_region_fpath,

                # The output directory of corresponding site models that should be used by the next step
                'output_sites_dpath': output_sites_dpath,

                # A single file that registers all of the sites writen to the output
                # site directory.
                'output_site_manifest_fpath': output_site_manifest_fpath,
            }
            filter_tracks.main(cmdline=cmdline, **filter_kwargs)

        except Exception:
            if config.skip_on_fail:
                print("WARNING: Exception occurred (printed below), passing input sites / region models as output")
                traceback.print_exception(*sys.exc_info())
                ingressed_assets['depth_filtered_sites'] = input_sites_dpath
                ingressed_assets['depth_filtered_regions'] = input_region_dpath
                shutil.copytree(input_sites_dpath, output_sites_dpath, dirs_exist_ok=True)
                shutil.copytree(input_region_dpath, output_region_dpath, dirs_exist_ok=True)
            else:
                raise
        else:
            # Validate and fix all outputs
            from geowatch.utils import util_framework
            util_framework.fixup_and_validate_site_and_region_models(
                region_dpath=output_region_fpath.parent,
                site_dpath=output_sites_dpath,
            )

    # HACK: Only egress sites and models if there was data to
    # generate them; with the intent of skipping the downstream
    # tasks in the DAG if there's no site output here
    ingressed_assets['depth_filtered_sites'] = output_sites_dpath
    ingressed_assets['depth_filtered_regions'] = output_region_dpath

    node_state.print_current_state(ingress_dir)

    # 4. Egress (envelop KWCOCO dataset in a STAC item and egress;
    #    will need to recursive copy the kwcoco output directory up to
    #    S3 bucket)
    print("* Egressing KWCOCO dataset and associated STAC item *")
    smartflow_egress(ingressed_assets,
                     local_region_path,
                     output_path,
                     outbucket,
                     aws_profile=aws_profile,
                     dryrun=dryrun,
                     newline=False)


if __name__ == "__main__":
    main()



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains functions to post-process ViP-DeepLab results."""

from typing import MutableMapping, Tuple

import numpy as np
import tensorflow as tf


def get_stitch_video_panoptic_prediction_info(
        concat_panoptic: np.ndarray,
        next_panoptic: np.ndarray,
        label_divisor: int,
        combine_offset: int = 2 ** 32) -> MutableMapping[int, int]:
    """Prepares the information for the stitching algorithm in ViP-DeepLab.

    Args:
      concat_panoptic: Panoptic prediction of the next frame by concatenating it
        with the current frame.
      next_panoptic: Panoptic prediction of the next frame.
      label_divisor: An integer specifying the label divisor of the dataset.
      combine_offset: An integer offset to combine concat and next panoptic.

    Returns:
      A map from the next frame instance IDs to the concatenated instances IDs.
    """

    def _ids_to_counts(id_array: np.ndarray):
        """Given a numpy array, a mapping from each entry to its count."""
        ids, counts = np.unique(id_array, return_counts=True)
        return dict(zip(ids, counts))

    # Pre-compute areas for all the segments.
    concat_segment_areas = _ids_to_counts(concat_panoptic)
    next_segment_areas = _ids_to_counts(next_panoptic)
    # Combine concat_panoptic and next_panoptic.
    intersection_id_array = (
            concat_panoptic.astype(np.int64) * combine_offset +
            next_panoptic.astype(np.int64))
    intersection_areas = _ids_to_counts(intersection_id_array)
    # Compute IoU and sort them.
    intersection_ious = []
    for intersection_id, intersection_area in intersection_areas.items():
        concat_panoptic_label = int(intersection_id // combine_offset)
        next_panoptic_label = int(intersection_id % combine_offset)
        concat_category_label = concat_panoptic_label // label_divisor
        next_category_label = next_panoptic_label // label_divisor
        if concat_category_label != next_category_label:
            continue
        concat_instance_label = concat_panoptic_label % label_divisor
        next_instance_label = next_panoptic_label % label_divisor
        # We skip 0 which is reserved for crowd.
        if concat_instance_label == 0 or next_instance_label == 0:
            continue
        union = (
                concat_segment_areas[concat_panoptic_label] +
                next_segment_areas[next_panoptic_label] - intersection_area)
        iou = intersection_area / union
        intersection_ious.append([concat_panoptic_label, next_panoptic_label, iou])
    intersection_ious = sorted(intersection_ious, key=lambda e: e[2])
    # Build mapping and inverse mapping. Two-way mapping guarantees 1-to-1
    # matching.
    map_concat_to_next = {}
    map_next_to_concat = {}
    for (concat_panoptic_label, next_panoptic_label, iou) in intersection_ious:
        map_concat_to_next[concat_panoptic_label] = next_panoptic_label
        map_next_to_concat[next_panoptic_label] = concat_panoptic_label

    remap_info = {
        next_label: concat_label
        for (concat_label, next_label) in map_concat_to_next.items()
        if map_next_to_concat[next_label] == concat_label
    }
    return remap_info


def stitch_video_panoptic_prediction(concat_panoptic: np.ndarray,
                                     next_panoptic: np.ndarray,
                                     label_divisor: int,
                                     overlap_offset: int = 128,
                                     combine_offset: int = 2 ** 32) -> np.ndarray:
    """The numpy implementation of the stitching algorithm in ViP-DeepLab.

    This function stitches a pair of image panoptic predictions to form video
    panoptic predictions by propagating instance IDs from concat_panoptic to
    next_panoptic based on IoU matching.

    Siyuan Qiao, Yukun Zhu, Hartwig Adam, Alan Yuille, and Liang-Chieh Chen.
    "ViP-DeepLab: Learning Visual Perception with Depth-aware Video Panoptic
    Segmentation." CVPR, 2021.

    Args:
      concat_panoptic: Panoptic prediction of the next frame by concatenating it
        with the current frame.
      next_panoptic: Panoptic prediction of the next frame.
      label_divisor: An integer specifying the label divisor of the dataset.
      overlap_offset: An integer offset to avoid overlap between the IDs in
        next_panoptic and the propagated IDs from concat_panoptic.
      combine_offset: An integer offset to combine concat and next panoptic.

    Returns:
      Panoptic prediction of the next frame with the instance IDs propragated
        from the concatenated panoptic prediction.
    """
    remap_info = get_stitch_video_panoptic_prediction_info(
        concat_panoptic, next_panoptic, label_divisor, combine_offset)

    new_panoptic = next_panoptic.copy()
    # Increase the panoptic instance ID to avoid overlap.
    new_category = new_panoptic // label_divisor
    new_instance = new_panoptic % label_divisor
    # We skip 0 which is reserved for crowd.
    instance_mask = new_instance > 0
    new_instance[instance_mask] = new_instance[instance_mask] + overlap_offset
    new_panoptic = new_category * label_divisor + new_instance
    # Match and propagate.
    for (next_panoptic_label, concat_panoptic_label) in remap_info.items():
        propagate_mask = next_panoptic == next_panoptic_label
        new_panoptic[propagate_mask] = concat_panoptic_label
    return new_panoptic


class VideoPanopticPredictionStitcher(tf.keras.Model):
    """The TF implementation of the stitching algorithm in ViP-DeepLab.

    It stitches a pair of image panoptic predictions to form video
    panoptic predictions by propagating instance IDs from concat_panoptic to
    next_panoptic based on IoU matching.

    Siyuan Qiao, Yukun Zhu, Hartwig Adam, Alan Yuille, and Liang-Chieh Chen.
    "ViP-DeepLab: Learning Visual Perception with Depth-aware Video Panoptic
    Segmentation." CVPR, 2021.
    """

    def __init__(self,
                 label_divisor: int,
                 combine_offset: int = 2 ** 32,
                 name: str = 'video_panoptic_prediction_stitcher'):
        """Initializes a TF video panoptic prediction stitcher.

        It also sets the overlap_offset to label_divisor // 2 to avoid overlap
        between IDs in next_panoptic and the propagated IDs from concat_panoptic.
        label_divisor // 2 gives equal space for the two frames.

        Args:
          label_divisor: An integer specifying the label divisor of the dataset.
          combine_offset: An integer offset to combine concat and next panoptic.
          name: A string specifying the model name.
        """
        super().__init__(name=name)
        self._label_divisor = label_divisor
        self._overlap_offset = label_divisor // 2
        self._combine_offset = combine_offset

    def _ids_to_counts(
            self, id_array: tf.Tensor) -> tf.lookup.experimental.MutableHashTable:
        """Given a tf Tensor, returns a mapping from its elements to their counts.

        Args:
          id_array: A tf.Tensor from which the function counts its elements.

        Returns:
          A MutableHashTable that maps the elements to their counts.
        """
        ids, _, counts = tf.unique_with_counts(tf.reshape(id_array, [-1]))
        table = tf.lookup.experimental.MutableHashTable(
            key_dtype=ids.dtype, value_dtype=counts.dtype, default_value=-1)
        table.insert(ids, counts)
        return table

    def _increase_instance_ids_by_offset(self, panoptic: tf.Tensor) -> tf.Tensor:
        """Increases instance IDs by self._overlap_offset.

        Args:
          panoptic: A tf.Tensor for the panoptic prediction.

        Returns:
          A tf.Tensor for paonptic prediction with increased instance ids.
        """
        category = panoptic // self._label_divisor
        instance = panoptic % self._label_divisor
        # We skip 0 which is reserved for crowd.
        instance_mask = tf.greater(instance, 0)
        tf.assert_less(
            tf.reduce_max(instance + self._overlap_offset),
            self._label_divisor,
            message='Any new instance IDs cannot exceed label_divisor.')
        instance = tf.where(instance_mask, instance + self._overlap_offset,
                            instance)
        return category * self._label_divisor + instance

    def _compute_and_sort_iou_between_panoptic_ids(
            self, panoptic_1: tf.Tensor,
            panoptic_2: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
        """Computes and sorts intersecting panoptic IDs by IoU.

        Args:
          panoptic_1: A tf.Tensor for the panoptic prediction for frame 1.
          panoptic_2: A tf.Tensor for the panoptic prediction for frame 2.

        Returns:
          A tuple of tf.Tensor storing the mapping ids between the two input frames
            with their IoUs in the ascending order.
        """
        segment_areas_1 = self._ids_to_counts(panoptic_1)
        segment_areas_2 = self._ids_to_counts(panoptic_2)
        intersection_id_array = (
                tf.cast(panoptic_1, tf.int64) * self._combine_offset +
                tf.cast(panoptic_2, tf.int64))
        intersection_areas_table = self._ids_to_counts(intersection_id_array)
        intersection_ids, intersection_areas = intersection_areas_table.export()
        panoptic_ids_1 = tf.cast(intersection_ids // self._combine_offset, tf.int32)
        panoptic_ids_2 = tf.cast(intersection_ids % self._combine_offset, tf.int32)
        category_ids_1 = panoptic_ids_1 // self._label_divisor
        category_ids_2 = panoptic_ids_2 // self._label_divisor
        instance_ids_1 = panoptic_ids_1 % self._label_divisor
        instance_ids_2 = panoptic_ids_2 % self._label_divisor
        unions = (
                segment_areas_1.lookup(panoptic_ids_1) +
                segment_areas_2.lookup(panoptic_ids_2) - intersection_areas)
        intersection_ious = intersection_areas / unions
        is_valid_intersection = tf.logical_and(
            tf.equal(category_ids_1, category_ids_2),
            tf.logical_and(
                tf.not_equal(instance_ids_1, 0), tf.not_equal(instance_ids_2, 0)))
        intersection_ious = tf.gather(intersection_ious,
                                      tf.where(is_valid_intersection)[:, 0])
        panoptic_ids_1 = tf.gather(panoptic_ids_1,
                                   tf.where(is_valid_intersection)[:, 0])
        panoptic_ids_2 = tf.gather(panoptic_ids_2,
                                   tf.where(is_valid_intersection)[:, 0])
        ious_indices = tf.argsort(intersection_ious)
        panoptic_ids_1 = tf.gather(panoptic_ids_1, ious_indices)
        panoptic_ids_2 = tf.gather(panoptic_ids_2, ious_indices)
        segment_areas_1.remove(segment_areas_1.export()[0])
        segment_areas_2.remove(segment_areas_2.export()[0])
        intersection_areas_table.remove(intersection_areas_table.export()[0])
        return panoptic_ids_1, panoptic_ids_2

    def _match_and_propagate_instance_ids(
            self, concat_panoptic: tf.Tensor, next_panoptic: tf.Tensor,
            concat_panoptic_ids: tf.Tensor,
            next_panoptic_ids: tf.Tensor) -> tf.Tensor:
        """Propagates instance ids based on instance id matching.

        It propagates the instance ids from concat_panoptic to next_panoptic based
        on the mapping specified by concat_panoptic_ids and next_panoptic_ids.

        Args:
          concat_panoptic: A tf.Tensor for the concat panoptic prediction.
          next_panoptic: A tf.Tensor for the next panoptic prediction.
          concat_panoptic_ids: A tf.Tensor for the matching ids in concat_panoptic.
          next_panoptic_ids: A tf.Tensor for the matching ids in next_panoptic.

        Returns:
          A tf.Tensor for the next panoptic prediction with instance ids propagated
            from concat_panoptic.
        """
        map_concat_to_next = tf.lookup.experimental.MutableHashTable(
            key_dtype=tf.int32, value_dtype=tf.int32, default_value=-1)
        map_concat_to_next.insert(
            tf.cast(concat_panoptic_ids, tf.int32),
            tf.cast(next_panoptic_ids, tf.int32))
        map_next_to_concat = tf.lookup.experimental.MutableHashTable(
            key_dtype=tf.int32, value_dtype=tf.int32, default_value=-1)
        map_next_to_concat.insert(
            tf.cast(next_panoptic_ids, tf.int32),
            tf.cast(concat_panoptic_ids, tf.int32))
        matched_concat_panoptic_ids, matched_next_panoptic_ids = (
            map_concat_to_next.export())
        returned_concat_panoptic_ids = map_next_to_concat.lookup(
            matched_next_panoptic_ids)
        matched_ids_mask = tf.equal(matched_concat_panoptic_ids,
                                    returned_concat_panoptic_ids)
        matched_concat_panoptic_ids = tf.gather(matched_concat_panoptic_ids,
                                                tf.where(matched_ids_mask)[:, 0])
        matched_next_panoptic_ids = tf.gather(matched_next_panoptic_ids,
                                              tf.where(matched_ids_mask)[:, 0])
        matched_concat_panoptic_ids = tf.expand_dims(
            tf.expand_dims(matched_concat_panoptic_ids, axis=-1), axis=-1)
        matched_next_panoptic_ids = tf.expand_dims(
            tf.expand_dims(matched_next_panoptic_ids, axis=-1), axis=-1)
        propagate_mask = tf.equal(next_panoptic, matched_next_panoptic_ids)
        panoptic_to_replace = tf.reduce_sum(
            tf.where(propagate_mask, matched_concat_panoptic_ids, 0),
            axis=0,
            keepdims=True)
        panoptic = tf.where(
            tf.reduce_any(propagate_mask, axis=0, keepdims=True),
            panoptic_to_replace, next_panoptic)
        panoptic = tf.ensure_shape(panoptic, next_panoptic.get_shape())
        map_concat_to_next.remove(map_concat_to_next.export()[0])
        map_next_to_concat.remove(map_next_to_concat.export()[0])
        return panoptic

    def call(self, concat_panoptic: tf.Tensor,
             next_panoptic: tf.Tensor) -> tf.Tensor:
        """Stitches the prediction from concat_panoptic and next_panoptic.

        Args:
          concat_panoptic: A tf.Tensor for the concat panoptic prediction.
          next_panoptic: A tf.Tensor for the next panoptic prediction.

        Returns:
          A tf.Tensor for the next panoptic prediction with instance ids propagated
            from concat_panoptic based on IoU matching.
        """
        next_panoptic = self._increase_instance_ids_by_offset(next_panoptic)
        concat_panoptic_ids, next_panoptic_ids = (
            self._compute_and_sort_iou_between_panoptic_ids(concat_panoptic,
                                                            next_panoptic))
        panoptic = self._match_and_propagate_instance_ids(concat_panoptic,
                                                          next_panoptic,
                                                          concat_panoptic_ids,
                                                          next_panoptic_ids)
        return panoptic



def patch_numpy_dtypes():
    import numpy as np
    np.bool = bool


def patch_numpy_2x():
    """
    Help compatability with 2.x

    This is mainly for handling dependencies that haven't officially ported to
    numpy 2.x yet. When they catch up this can be removed.

    Main culprits are tensorboard and lightning

    Ignore:
        from geowatch.monkey.monkey_numpy import patch_numpy_2x
        patch_numpy_2x()
    """
    import numpy as np
    if np.lib.NumpyVersion(np.__version__) >= '2.0.0b1':
        np.string_ = np.bytes_
        np.unicode_ = np.str_
        np.Inf = np.inf



"""
Follows the patches submitted in https://github.com/omni-us/jsonargparse/pull/532

Refactor References:
    ~/code/jsonargparse/jsonargparse/_signatures.py
    ~/code/jsonargparse/jsonargparse/_parameter_resolvers.py

"""
import dataclasses
import inspect
from typing import Any, Callable, List, Optional, Set, Tuple, Type, Union
# from jsonargparse._common import LoggerProperty
from jsonargparse._parameter_resolvers import get_signature_parameters
from jsonargparse._parameter_resolvers import get_parameter_origins
from jsonargparse._parameter_resolvers import ParamData
from jsonargparse._optionals import get_doc_short_description
from jsonargparse._signatures import SignatureArguments
from argparse import ArgumentParser
from jsonargparse._common import LoggerProperty   # NOQA
from jsonargparse._common import get_class_instantiator   # NOQA
from jsonargparse._common import get_generic_origin   # NOQA
from jsonargparse._common import get_unaliased_type   # NOQA
from jsonargparse._common import is_dataclass_like   # NOQA
from jsonargparse._common import is_subclass   # NOQA
from jsonargparse._typehints import ActionTypeHint
from jsonargparse._typehints import is_optional
from jsonargparse.typing import register_pydantic_type

kinds = inspect._ParameterKind
inspect_empty = inspect._empty


@dataclasses.dataclass
class ParamData_Extension(ParamData):
    name: str
    annotation: Any
    default: Any = inspect._empty
    kind: Optional[inspect._ParameterKind] = None
    doc: Optional[str] = None
    component: Optional[Union[Callable, Type, Tuple]] = None
    parent: Optional[Union[Type, Tuple]] = None
    origin: Optional[Union[str, Tuple]] = None
    short_aliases: Optional[List[str]] = None
    long_aliases: Optional[List[str]] = None

    def _resolve_args_and_dest(self, is_required=False, as_positional=False, nested_key: Optional[str] = None):
        name = self.name
        dest = (nested_key + "." if nested_key else "") + name
        if is_required and as_positional:
            args = [dest]
        else:
            long_names = [name] + list((self.long_aliases or []))
            short_names = list(self.short_aliases or [])
            nest_prefix = nested_key + "." if nested_key else ""
            short_option_strings = ["-" + nest_prefix + n for n in short_names]
            long_option_strings = ["--" + nest_prefix + n for n in long_names]
            args = short_option_strings + long_option_strings
        return args, dest


class SignatureArguments_Extension(SignatureArguments):

    def _add_signature_arguments(
        self,
        function_or_class,
        method_name,
        nested_key: Optional[str],
        as_group: bool = True,
        as_positional: bool = False,
        skip: Optional[Set[Union[str, int]]] = None,
        fail_untyped: bool = True,
        sub_configs: bool = False,
        instantiate: bool = True,
        linked_targets: Optional[Set[str]] = None,
    ) -> List[str]:
        """Adds arguments from parameters of objects based on signatures and docstrings.

        Args:
            function_or_class: Object from which to add arguments.
            method_name: Class method from which to add arguments.
            nested_key: Key for nested namespace.
            as_group: Whether arguments should be added to a new argument group.
            as_positional: Whether to add required parameters as positional arguments.
            skip: Names of parameters or number of positionals that should be skipped.
            fail_untyped: Whether to raise exception if a required parameter does not have a type.
            sub_configs: Whether subclass type hints should be loadable from inner config file.
            instantiate: Whether the class group should be instantiated by :code:`instantiate_classes`.

        Returns:
            The list of arguments added.

        Raises:
            ValueError: When there are required parameters without at least one valid type.
        """
        params = get_signature_parameters(function_or_class, method_name, logger=self.logger)

        skip_positionals = [s for s in (skip or []) if isinstance(s, int)]
        if skip_positionals:
            if len(skip_positionals) > 1 or any(p <= 0 for p in skip_positionals):
                raise ValueError(f"Unexpected number of positionals to skip: {skip_positionals}")
            names = {p.name for p in params[: skip_positionals[0]]}
            params = params[skip_positionals[0] :]
            self.logger.debug(
                f"Skipping parameters {names} because {skip_positionals[0]} positionals requested to be skipped."
            )

        prefix = "--" + (nested_key + "." if nested_key else "")
        for param in params:
            if skip and param.name in skip:
                continue
            if prefix + param.name in self._option_string_actions:  # type: ignore[attr-defined]
                raise ValueError(
                    f"Unable to add parameter '{param.name}' from {function_or_class} because "
                    f"argument '{prefix + param.name}' already exists."
                )

        ## Create group if requested ##
        doc_group = get_doc_short_description(function_or_class, method_name, logger=self.logger)
        component = getattr(function_or_class, method_name) if method_name else function_or_class
        container = self._create_group_if_requested(
            component,
            nested_key,
            as_group,
            doc_group,
            config_load=len(params) > 0,
            instantiate=instantiate,
        )

        ## Add parameter arguments ##
        added_args: List[str] = []

        for param in params:
            self._add_signature_parameter(
                container,
                nested_key,
                param,
                added_args,
                skip={s for s in (skip or []) if isinstance(s, str)},
                fail_untyped=fail_untyped,
                sub_configs=sub_configs,
                linked_targets=linked_targets,
                as_positional=as_positional,
            )

        if hasattr(function_or_class, "__scriptconfig__"):
            # Integrate with scriptconfig style classes.
            # When a function/class has a __scriptconfig__ object that means it
            # should accept any of the options defined in the config as keyword
            # arguments. We can utilize additional metadata stored in the
            # scriptconfig object to enrich our CLI.
            config_cls = function_or_class.__scriptconfig__
            self._add_scriptconfig_arguments(
                config_cls=config_cls,
                added_args=added_args,
                function_or_class=function_or_class,
                component=component,
                container=container,
                nested_key=nested_key,
                fail_untyped=fail_untyped,
                as_positional=as_positional,
                sub_configs=sub_configs,
                skip=skip,
                linked_targets=linked_targets,
            )

        return added_args

    def _add_scriptconfig_arguments(
        self,
        config_cls,
        added_args,
        function_or_class,
        component,
        container,
        nested_key,
        fail_untyped: bool = True,
        as_positional: bool = False,
        sub_configs: bool = False,
        skip: Optional[Set[Union[str, int]]] = None,
        linked_targets: Optional[Set[str]] = None,
    ):

        for key, value in config_cls.__default__.items():
            from scriptconfig import value as value_mod

            if not isinstance(value, value_mod.Value):
                # hack
                value = value_mod.Value(value)
            type_ = value.parsekw["type"]
            if type_ is None or not isinstance(type_, type):
                annotation = inspect._empty
            else:
                annotation = type_
            param = ParamData_Extension(
                name=key,
                annotation=annotation,
                kind=inspect.Parameter.KEYWORD_ONLY,
                default=value.value,
                doc=value.parsekw["help"],
                component=component,
                parent=function_or_class,
                short_aliases=value.short_alias,
                long_aliases=value.alias,
            )
            self._add_signature_parameter(
                container,
                nested_key,
                param,
                added_args,
                skip={s for s in (skip or []) if isinstance(s, str)},
                fail_untyped=fail_untyped,
                sub_configs=sub_configs,
                linked_targets=linked_targets,
                as_positional=as_positional,
            )

    def _add_signature_parameter(
        self,
        container,
        nested_key: Optional[str],
        param,
        added_args: List[str],
        skip: Optional[Set[str]] = None,
        fail_untyped: bool = True,
        as_positional: bool = False,
        sub_configs: bool = False,
        instantiate: bool = True,
        linked_targets: Optional[Set[str]] = None,
        default: Any = inspect_empty,
        **kwargs,
    ):
        name = param.name
        kind = param.kind
        annotation = param.annotation
        if default == inspect_empty:
            default = param.default
            if default == inspect_empty and is_optional(annotation):
                default = None
        is_required = default == inspect_empty
        src = get_parameter_origins(param.component, param.parent)
        skip_message = f'Skipping parameter "{name}" from "{src}" because of: '
        if not fail_untyped and annotation == inspect_empty:
            annotation = Any
            default = None if is_required else default
            is_required = False
        if is_required and linked_targets is not None and name in linked_targets:
            default = None
            is_required = False
        if (
            kind in {kinds.VAR_POSITIONAL, kinds.VAR_KEYWORD}
            or (not is_required and name[0] == "_")
            or (annotation == inspect_empty and not is_required and default is None)
        ):
            return
        elif skip and name in skip:
            self.logger.debug(skip_message + "Parameter requested to be skipped.")
            return
        if is_factory_class(default):
            default = param.parent.__dataclass_fields__[name].default_factory()
        if annotation == inspect_empty and not is_required:
            annotation = Union[type(default), Any]
        if "help" not in kwargs:
            kwargs["help"] = param.doc
        if not is_required:
            kwargs["default"] = default
            if default is None and not is_optional(annotation, object):
                annotation = Optional[annotation]
        elif not as_positional:
            kwargs["required"] = True
        is_subclass_typehint = False
        is_dataclass_like_typehint = is_dataclass_like(annotation)
        # dest = (nested_key + "." if nested_key else "") + name
        # args = [dest if is_required and as_positional else "--" + dest]
        args, dest = param._resolve_args_and_dest(is_required, as_positional, nested_key)
        if param.origin:
            parser = container
            if not isinstance(container, ArgumentParser):
                parser = getattr(container, "parser")
            group_name = "; ".join(str(o) for o in param.origin)
            if group_name in parser.groups:
                container = parser.groups[group_name]
            else:
                container = parser.add_argument_group(
                    f"Conditional arguments [origins: {group_name}]",
                    name=group_name,
                )
        if (
            annotation in {str, int, float, bool}
            or is_subclass(annotation, (str, int, float))
            or is_dataclass_like_typehint
        ):
            kwargs["type"] = annotation
            register_pydantic_type(annotation)
        elif annotation != inspect_empty:
            try:
                is_subclass_typehint = ActionTypeHint.is_subclass_typehint(annotation, all_subtypes=False)
                kwargs["type"] = annotation
                sub_add_kwargs: dict = {"fail_untyped": fail_untyped, "sub_configs": sub_configs}
                if is_subclass_typehint:
                    prefix = f"{name}.init_args."
                    subclass_skip = {s[len(prefix) :] for s in skip or [] if s.startswith(prefix)}
                    sub_add_kwargs["skip"] = subclass_skip
                else:
                    register_pydantic_type(annotation)
                enable_path = sub_configs and (
                    is_subclass_typehint or ActionTypeHint.is_return_subclass_typehint(annotation)
                )
                args = ActionTypeHint.prepare_add_argument(
                    args=args,
                    kwargs=kwargs,
                    enable_path=enable_path,
                    container=container,
                    logger=self.logger,
                    sub_add_kwargs=sub_add_kwargs,
                )
            except ValueError as ex:
                self.logger.debug(skip_message + str(ex))
        if "type" in kwargs or "action" in kwargs:
            sub_add_kwargs = {
                "fail_untyped": fail_untyped,
                "sub_configs": sub_configs,
                "instantiate": instantiate,
            }
            if is_dataclass_like_typehint:
                kwargs.update(sub_add_kwargs)
            with ActionTypeHint.allow_default_instance_context():
                action = container.add_argument(*args, **kwargs)
            action.sub_add_kwargs = sub_add_kwargs
            if is_subclass_typehint and len(subclass_skip) > 0:
                action.sub_add_kwargs["skip"] = subclass_skip
            added_args.append(dest)
        elif is_required and fail_untyped:
            raise ValueError(
                "With fail_untyped=True, all mandatory parameters must have a supported"
                f" type. Parameter '{name}' from '{src}' does not specify a type."
            )


def is_factory_class(value):
    return value.__class__ == dataclasses._HAS_DEFAULT_FACTORY_CLASS


def raise_unexpected_value(message: str, val: Any = inspect._empty, exception: Optional[Exception] = None):
    """
    Patch to disable overzelous type hints
    """
    if val is not inspect._empty:
        message += f". Got value: {val} (note: jsonargparse warnings can be overzealous)"
    import warnings
    warnings.warn(message)
    # raise ValueError(message) from exception


def apply_monkeypatch():
    """
    Dynamically add the monkeypatch to jsonargparse
    """
    import jsonargparse
    # for key in jsonargparse.__all__:
    #     value = getattr(jsonargparse, key)
    #     if isinstance(value, type) and issubclass(value, SignatureArguments):

    # Hack in the new methods from SignatureArguments
    jsonargparse._signatures.ParamData = ParamData_Extension
    jsonargparse._parameter_resolvers.ParamData = ParamData_Extension
    jsonargparse._signatures.SignatureArguments._add_signature_arguments = SignatureArguments_Extension._add_signature_arguments
    jsonargparse._signatures.SignatureArguments._add_signature_parameter = SignatureArguments_Extension._add_signature_parameter
    jsonargparse._signatures.SignatureArguments._add_scriptconfig_arguments = SignatureArguments_Extension._add_scriptconfig_arguments

    # hack to disable config checks
    import ubelt as ub
    jsonargparse._core.ArgumentParser.check_config = ub.identity
    jsonargparse._typehints.raise_unexpected_value = raise_unexpected_value



import pathlib
import setuptools

# The directory containing this file
HERE = pathlib.Path(__file__).parent
# The text of the README file
README = (HERE / "README.md").read_text()
# This call to setup() does all the work
setuptools.setup(
    name="lydorn_utils",
    version="0.0.1",
    description="Various utilities for deep learning projects",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Nicolas Girard",
    author_email="nicolas.jp.girard@gmail.com",
    license="INRIA",
    classifiers=[
        "License :: INRIA",
        "Programming Language :: Python"
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.6"
)
